// ========================================================================================== //
//             ___  _   _    __    __  __    __    ____  __    ____  _____  _  _              //
//            / __)( )_( )  /__\  (  \/  )  /__\  ( ___)(  )  ( ___)(  _  )( \( )             //
//           ( (__  ) _ (  /(__)\  )    (  /(__)\  )__)  )(__  )__)  )(_)(  )  (              //
//            \___)(_) (_)(__)(__)(_/\/\_)(__)(__)(____)(____)(____)(_____)(_)\_)             //
//            -------------------------------------------- G A M E   E N G I N E              //
//                                                                                            //
// ========================================================================================== //
// Developed and implemented by:                                                              //
//     Oskam Thomas                                                                           //
//                                                                                            //
//                                                                                            //
// Copyright 2007/2008. All rights reserved.                                                  //
// ========================================================================================== //




// ============================================================================================
// ---- Shader Parameter ----------------------------------------------------------------------
// ============================================================================================


// view projection matrix
float4x4 xViewProjection;


// translation in world coordinates that is applied to the whole heightfield
float3 xHeightfieldTranslation;


// factor with which the heightmap height values are multiplied, in world coordinates
float xMaxAlteringHeight;


// factor multiplied to the index (x,y number) of each instanced patch, to get the texture 
// coords on the heightmap 
float2 xHeightfieldPixelOffset;


// world size of a patch (patch is quadratic)
float xPatchWidth;


// texture coords stretch factor applied to the local texture coords, one component per texture channel
float4 localTextureStretchFactor;

// factore to be multiplied to y texture coordinate to restore quadratic relation (for local texturing)
float localYTextureRelation;


// offset in heightmap texture coords between the individual pixel of the heightmap
float2 xTexCoordDelta;


// offset between heighmap pixel in world coordinates
float xTexCoordWorldDelta;


// global light direction
float3 xLightDirection;


// terrain alpha blending parameter, x = start dist, y = end dist
float2 xAlpha_Start_End;


// camera position
float3 xCameraPosition;








// ============================================================================================
// ---- Internal VertexSh-PixelSh parameter definitions ---------------------------------------
// ============================================================================================




// normal Vertex Shader Input
struct VS_input {
	float4 Position		: POSITION;
	float2 TexCoords	: TEXCOORD0;
};


// instanced Vertex Shader Input
struct VS_instInput {
	float4 Position		: POSITION;
	float2 TexCoords	: TEXCOORD0;
	float2 PatchIndex	: TEXCOORD2;
};


// normal Pixel Shader input
struct PS_input {
    float4 Position			: POSITION; 
    float3 Normal			: NORMAL; 
	float2 GlobalTexCoords	: TEXCOORD0;
	float2 LocalTexCoordsR	: TEXCOORD1;
	float2 LocalTexCoordsG	: TEXCOORD2;
	float2 LocalTexCoordsB	: TEXCOORD3;
	float2 LocalTexCoordsA	: TEXCOORD4;
	float  Distance			: TEXCOORD5;
};


// normal Frame Buffer input
struct FB_input {
	float4 Color		: COLOR0;
};



// constants:
float2 textureSize = {256, 256}; // this value is set on purpose to the half resolution, in order to achieve better mipmapping results





// ============================================================================================
// ---- Textures and Texture Sampler ----------------------------------------------------------
// ============================================================================================



// GLOBAL TEXTURES: each pixel defines a value for the grid
//=============================================================================================

// heightfield texture: used to perturbate the height of the grid
uniform Texture xHeightfield;
sampler HeightTextureSampler = sampler_state
{
	texture = <xHeightfield>;
	//magfilter = LINEAR;
	//minfilter = LINEAR;
	//mipfilter = LINEAR;
	magfilter = NONE;
	minfilter = NONE;
	mipfilter = NONE;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};


sampler SmoothHeightTextureSampler = sampler_state
{
	texture = <xHeightfield>;
	magfilter = LINEAR;
	minfilter = LINEAR;
	//mipfilter = LINEAR;
	//magfilter = NONE;
	//minfilter = NONE;
	mipfilter = NONE;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};


// channel texture: each color channel defines an intensity value for a local texture
uniform Texture xChannelfield;
sampler ChannelTextureSampler = sampler_state
{
	texture = <xChannelfield>;
	//magfilter = LINEAR;
	//minfilter = LINEAR;
	//mipfilter = LINEAR;
	//magfilter = NONE;
	//minfilter = NONE;
	//mipfilter = NONE;
	magfilter = Anisotropic;
	minfilter = Anisotropic;
	maxanisotropy = 16;
	mipfilter = LINEAR;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};


// light texture: each pixel contains the light value received by that position of the heightmap
uniform Texture xNormalTexture;
sampler NormalTextureSampler = sampler_state
{
	texture = <xNormalTexture>;
	magfilter = LINEAR;
	minfilter = LINEAR;
	mipfilter = LINEAR;
	//magfilter = NONE;
	//minfilter = NONE;
	//mipfilter = NONE;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};



// LOCAL (CHANNEL) TEXTURES: textures for the final color (selected using channel texture
//=============================================================================================


// red texture channel
uniform Texture xRedChannelTexture;
sampler RedChannelTextureSampler = sampler_state
{
	texture = <xRedChannelTexture>;
	magfilter = Anisotropic;
	minfilter = Anisotropic;
	maxanisotropy = 16;
	mipfilter = LINEAR;
	//magfilter = NONE;
	//minfilter = NONE;
	//mipfilter = NONE;
	AddressU  = wrap;
	AddressV  = wrap;
};


// green texture channel
uniform Texture xGreenChannelTexture;
sampler GreenChannelTextureSampler = sampler_state
{
	texture = <xGreenChannelTexture>;
	magfilter = Anisotropic;
	minfilter = Anisotropic;
	maxanisotropy = 16;
	mipfilter = LINEAR;
	//magfilter = NONE;
	//minfilter = NONE;
	//mipfilter = NONE;
	AddressU  = wrap;
	AddressV  = wrap;
};


// blue texture channel
uniform Texture xBlueChannelTexture;
sampler BlueChannelTextureSampler = sampler_state
{
	texture = <xBlueChannelTexture>;
	magfilter = Anisotropic;
	minfilter = Anisotropic;
	maxanisotropy = 16;
	mipfilter = LINEAR;
	//magfilter = NONE;
	//minfilter = NONE;
	//mipfilter = NONE;
	AddressU  = wrap;
	AddressV  = wrap;
};


// alpha texture channel
uniform Texture xAlphaChannelTexture;
sampler AlphaChannelTextureSampler = sampler_state
{
	texture = <xAlphaChannelTexture>;
	magfilter = Anisotropic;
	minfilter = Anisotropic;
	maxanisotropy = 16;
	mipfilter = LINEAR;
	//magfilter = NONE;
	//minfilter = NONE;
	//mipfilter = NONE;
	AddressU  = wrap;
	AddressV  = wrap;
};





// ============================================================================================
// ---- Vertex Shader  ------------------------------------------------------------------------
// ============================================================================================


// Instanced Vertex Shader, NO computation of NORMAL
PS_input Instanced_Heightfield_VS( VS_input input, float2 PatchIndex : TEXCOORD2 )
{
	// 0) declare Output and initialize all variables to 0
	PS_input Output = (PS_input)0;
	
	// 2) get heightfield texture value & compute height value
	float2 globalTexCoords = input.TexCoords + PatchIndex * xHeightfieldPixelOffset;
	float texColor = tex2Dlod(HeightTextureSampler, float4(globalTexCoords, 0, 0)).x;
	float3 h = {0.0f, xMaxAlteringHeight * texColor, 0.0f};
	
	// 3) compute translation
	float3 t = {PatchIndex.x * xPatchWidth, 0.0f, PatchIndex.y * xPatchWidth};

	
	// 4) store output
	float4 factor = {1.0, 1.0, 1.0, 1.0};
	float4 finalPos4 = float4(input.Position + h + t + xHeightfieldTranslation, 1.0f);
	Output.Position = mul(finalPos4, xViewProjection);
	Output.GlobalTexCoords = globalTexCoords;
	
	float2 LocalTexCoords = (globalTexCoords) * float2(1, localYTextureRelation);
	Output.LocalTexCoordsR = LocalTexCoords * float2(localTextureStretchFactor.x, localTextureStretchFactor.x);
	Output.LocalTexCoordsG = LocalTexCoords * float2(localTextureStretchFactor.y, localTextureStretchFactor.y);
	Output.LocalTexCoordsB = LocalTexCoords * float2(localTextureStretchFactor.z, localTextureStretchFactor.z);
	Output.LocalTexCoordsA = LocalTexCoords * float2(localTextureStretchFactor.w, localTextureStretchFactor.w);
	
	float4 normalColor = tex2Dlod(NormalTextureSampler, float4(globalTexCoords,0.0f,1.0f));
	Output.Normal = normalize(normalColor.xyz);
	
	// 5) compute distance to camera for alpha distance blending
	Output.Distance = length(xCameraPosition.xz - finalPos4.xz);
	
	return Output;
}






// Instanced Shader that computes normal
PS_input Normaled_Instanced_Heightfield_VS( VS_input input, float2 PatchIndex : TEXCOORD2 )
{
	// 0) declare Output and initialize all variables to 0
	PS_input Output = (PS_input)0;
	
	// 2) get heightfield texture value & compute height value
	float2 globalTexCoords = input.TexCoords + PatchIndex * xHeightfieldPixelOffset;
	float texColor = tex2Dlod(HeightTextureSampler, float4(globalTexCoords,0.0f,1.0f)).x;
	float3 h = {0.0f, xMaxAlteringHeight * texColor, 0.0f};
	
	// 3) compute translation
	float3 t = {PatchIndex.x * xPatchWidth, 0.0f, PatchIndex.y * xPatchWidth};

	
	// 4) store output
	float4 finalPos4 = float4(input.Position + h + t + xHeightfieldTranslation, 1.0f);
	Output.Position = mul(finalPos4, xViewProjection);
	Output.GlobalTexCoords = globalTexCoords;
	
	float2 LocalTexCoords = (globalTexCoords) * float2(1, localYTextureRelation);
	Output.LocalTexCoordsR = LocalTexCoords * float2(localTextureStretchFactor.x, localTextureStretchFactor.x);
	Output.LocalTexCoordsG = LocalTexCoords * float2(localTextureStretchFactor.y, localTextureStretchFactor.y);
	Output.LocalTexCoordsB = LocalTexCoords * float2(localTextureStretchFactor.z, localTextureStretchFactor.z);
	Output.LocalTexCoordsA = LocalTexCoords * float2(localTextureStretchFactor.w, localTextureStretchFactor.w);
	
	// 5) compute distance to camera for alpha distance blending
	Output.Distance = length(xCameraPosition.xz - finalPos4.xz);
	
	
	// 6) dynamically compute normal
	
		// 6a) get height values from neighbors
		float2 texCoordOffsetVec = xTexCoordDelta * 2.0f;
		float north_h = tex2Dlod(HeightTextureSampler, float4(globalTexCoords + float2(0.0f, xTexCoordDelta.y),  0.0f, 1.0f)).x;
		float south_h = tex2Dlod(HeightTextureSampler, float4(globalTexCoords + float2(0.0f, -xTexCoordDelta.y ), 0.0f, 1.0f)).x;
		float east_h =  tex2Dlod(HeightTextureSampler, float4(globalTexCoords + float2(xTexCoordDelta.x, 0.0f),  0.0f, 1.0f)).x;
		float west_h =  tex2Dlod(HeightTextureSampler, float4(globalTexCoords + float2(-xTexCoordDelta.x, 0.0f), 0.0f, 1.0f)).x;

		// 6b) define parameters
		float two_texDelta = xTexCoordWorldDelta + xTexCoordWorldDelta;
		float divisor = 1.0f / two_texDelta;

		// 6c) compute normal 
		float3 south_minus_north = {0.0f, (south_h - north_h)*xMaxAlteringHeight, -(two_texDelta)};
		float3 east_minus_west = {-(two_texDelta), (east_h - west_h)*xMaxAlteringHeight, 0.0f};
		Output.Normal = normalize(cross((south_minus_north)*divisor, (east_minus_west)*divisor));


	
	return Output;
}








//------------------------------------------------------------------------------------------
// Helper Function: computes mipmap level --------------------------------------------------
//------------------------------------------------------------------------------------------

float MipmapLevel(float2 uv, float2 textureSize)
{
	float2 dx = ddx(uv * textureSize.x);
	float2 dy = ddy(uv * textureSize.y);
	float d = max( dot(dx, dx), dot(dy, dy) );

	//return log2( sqrt(d) );
	return 0.5 * log2( d );
}









// ============================================================================================
// ---- Pixel Shader  -------------------------------------------------------------------------
// ============================================================================================



float4 computeGroundColor(float2 GlobalTexCoords, float2 LocalTexCoordsR, float2 LocalTexCoordsG, float2 LocalTexCoordsB, float2 LocalTexCoordsA) {
	
	// pull color channel code
	float4 channelColor = tex2D( ChannelTextureSampler, GlobalTexCoords );

	// get color
	float lodR = MipmapLevel(LocalTexCoordsR, textureSize);
	float lodG = MipmapLevel(LocalTexCoordsG, textureSize);
	float lodB = MipmapLevel(LocalTexCoordsB, textureSize);
	float lodA = MipmapLevel(LocalTexCoordsA, textureSize);
	float4 rTexColor = tex2Dlod(RedChannelTextureSampler, float4(LocalTexCoordsR,0.0f,lodR));
	float4 gTexColor = tex2Dlod(GreenChannelTextureSampler, float4(LocalTexCoordsG,0.0f,lodG));
	float4 bTexColor = tex2Dlod(BlueChannelTextureSampler, float4(LocalTexCoordsB,0.0f,lodB));
	float4 aTexColor = tex2Dlod(AlphaChannelTextureSampler, float4(LocalTexCoordsA,0.0f,lodA));
	
	return ((channelColor.x * rTexColor) + (channelColor.y * gTexColor) + (channelColor.z * bTexColor) + (channelColor.w * aTexColor));
}







// Pixel Shader that computes lighting using xLightDirection as global light
FB_input Lit_Heightfield_PS(PS_input input)
{
	//0) declare output and intialize with 0
	FB_input Output = (FB_input)0;
	
	// 1) compute lighting intesity
	float lightingFactor = max(0.35 ,min(0.65f, dot(input.Normal, -xLightDirection))) + 0.35;
	
	
	// 2) define output color
	Output.Color = computeGroundColor(input.GlobalTexCoords, 
	                                  input.LocalTexCoordsR, 
	                                  input.LocalTexCoordsG, 
	                                  input.LocalTexCoordsB, 
	                                  input.LocalTexCoordsA) * lightingFactor;
	
	// 3) make alpha blending dependant on distance
	Output.Color.w =  lerp(1, 0, input.Distance-xAlpha_Start_End.x)/(xAlpha_Start_End.y-xAlpha_Start_End.x);

	return Output;
}





// ============================================================================================
// ---- Shader Techniques  --------------------------------------------------------------------
// ============================================================================================

technique instancedHeightfielded
{
pass
    {
        VertexShader = compile vs_3_0 Instanced_Heightfield_VS();
        PixelShader = compile ps_3_0 Lit_Heightfield_PS();
    }
}



technique instancedHeightfieldedComputeNormals
{
pass
    {
        VertexShader = compile vs_3_0 Normaled_Instanced_Heightfield_VS();
        PixelShader = compile ps_3_0 Lit_Heightfield_PS();
    }
}

