// ========================================================================================== //
//             ___  _   _    __    __  __    __    ____  __    ____  _____  _  _              //
//            / __)( )_( )  /__\  (  \/  )  /__\  ( ___)(  )  ( ___)(  _  )( \( )             //
//           ( (__  ) _ (  /(__)\  )    (  /(__)\  )__)  )(__  )__)  )(_)(  )  (              //
//            \___)(_) (_)(__)(__)(_/\/\_)(__)(__)(____)(____)(____)(_____)(_)\_)             //
//            -------------------------------------------- G A M E   E N G I N E              //
//                                                                                            //
// ========================================================================================== //
// Developed and implemented by:                                                              //
//     Mader Theodor                                                                          //
//     Oskam Thomas                                                                           //
//                                                                                            //
// Copyright 2007-2009. All rights reserved.                                                  //
// ========================================================================================== //

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;

namespace HFShader.Camera
{
    /// <summary>
    /// The class CamConfiguration encapsulates a camera state containing a position
    /// and an orientation
    /// </summary>
    public class CamConfiguration
    {

        private Vector3 mPosition;
        public Vector3 position
        {
            get { return mPosition; }
            set { mPosition = value;
                if (float.IsNaN(mPosition.X))
                {
                    float error = 1;
                    error++;
                }
            }
        }

        // rotation around the up vector
        private float mPitch;
        public float pitch
        {
            get { return mPitch; }
            set
            {
                mPitch = value;
                while (mPitch > 360.0f) mPitch -= 360.0f;
                while (mPitch < 0.0f) mPitch += 360.0f;
                if (float.IsNaN(mPitch))
                {
                    float error = 1;
                    error++;
                }
            }
        }

        // rotation around the right vector
        private float mYaw;
        public float yaw
        {
            get { return mYaw; }
            set
            {
                mYaw = value;
                while (mYaw > 360.0f) mYaw -= 360.0f;
                if (float.IsNaN(mYaw))
                {
                    float error = 1;
                    error++;
                }
            }
        }

        // rotation around the dir vector
        private float mRoll;
        public float roll
        {
            get { return mRoll; }
            set
            {
                mRoll = value;
                while (mRoll > 360.0f) mRoll -= 360.0f;
                if (float.IsNaN(mRoll)) {
                    float error = 1;
                    error++;
                }
            }
        }


        // constructor
        public CamConfiguration()
        {
            mPosition = Vector3.Zero;
            mPitch = 0.0f;
            mRoll = 0.0f;
            mYaw = 0.0f;
        }

        // set vector for direction
        public void setDirection(Vector3 direction)
        {
            Vector3 lookDir = direction;

            if (lookDir.Length() < 0.001)
            {
                // pitch and yaw will still be valid
                // so we simply return
                return;
            }

            lookDir.Normalize();


            if (Math.Abs(lookDir.Y) > 0.999f) // look dir = +/- unitY = up or down
            {
                mPitch = 0;

                float yawAngle = (float)Math.Acos(0.0f);
                if (lookDir.Y <= 0) // right
                {
                    yawAngle = -yawAngle;
                }
                mYaw = MathHelper.ToDegrees(yawAngle);
            }
            else
            { // lookDir.Y < 1

                Vector3 lookDirXZ = new Vector3(lookDir.X, 0.0f, lookDir.Z);
                lookDirXZ.Normalize();

                // compute desired pitch
                float pitchAngle = (float)Math.Acos(Vector3.Dot(lookDirXZ, new Vector3(1.0f, 0.0f, 0.0f)));
                if (lookDirXZ.Z >= 0) // right
                {
                    pitchAngle = 2 * (float)Math.PI - pitchAngle;
                }
                mPitch = MathHelper.ToDegrees(pitchAngle);


                // compute desired yaw
                // float v = (float)Math.Acos(Vector3.Dot(lookDirXZ, lookDir));
                float v = Vector3.Dot(lookDirXZ, lookDir);
                if (v > 1.0f)
                {
                    v = 1.0f; // oh this one is cool: 
                    // v can be > 1, although both vectors are normed... 
                }


                float yawAngle = (float)Math.Acos(v);
                if (lookDir.Y <= 0) // right
                {
                    yawAngle = 2.0f * (float)Math.PI - yawAngle;
                }
                mYaw = MathHelper.ToDegrees(yawAngle);
            }
        }



        public Vector3 getDirection()
        {

            Matrix pitchMatrix = Matrix.CreateRotationY(MathHelper.ToRadians(mPitch));
            Matrix yawMatrix = Matrix.CreateRotationZ(MathHelper.ToRadians(mYaw));

            Matrix transformation = yawMatrix * pitchMatrix;
            Vector3 lookDir = Vector3.UnitX;
            lookDir = Vector3.Transform(lookDir, transformation);
            lookDir.Normalize();
            return lookDir;
        }




        public BoundingFrustum getCamFrustum(Matrix camProjection)
        {
            return new BoundingFrustum(createModelViewMatrix() * camProjection);
        }

        public Matrix createModelViewMatrix()
        {
            // create modelview matrix
            Matrix rollMatrix = Matrix.CreateRotationX(MathHelper.ToRadians(this.roll));
            Matrix pitchMatrix = Matrix.CreateRotationY(MathHelper.ToRadians(this.pitch));
            Matrix yawMatrix = Matrix.CreateRotationZ(MathHelper.ToRadians(this.yaw));

            Matrix transformation = rollMatrix * yawMatrix * pitchMatrix;

            Vector3 up = Vector3.Transform(Vector3.UnitY, rollMatrix * yawMatrix * pitchMatrix);

            Vector3 lookDir = Vector3.UnitX;
            lookDir = Vector3.Transform(lookDir, transformation);
            lookDir.Normalize();

            return Matrix.CreateLookAt(this.position, this.position + lookDir, up);
        }

    }
}
