// ========================================================================================== //
//             ___  _   _    __    __  __    __    ____  __    ____  _____  _  _              //
//            / __)( )_( )  /__\  (  \/  )  /__\  ( ___)(  )  ( ___)(  _  )( \( )             //
//           ( (__  ) _ (  /(__)\  )    (  /(__)\  )__)  )(__  )__)  )(_)(  )  (              //
//            \___)(_) (_)(__)(__)(_/\/\_)(__)(__)(____)(____)(____)(_____)(_)\_)             //
//            -------------------------------------------- G A M E   E N G I N E              //
//                                                                                            //
// ========================================================================================== //
// Developed and implemented by:                                                              //
//     Mader Theodor                                                                          //
//     Oskam Thomas                                                                           //
//                                                                                            //
// Copyright 2007-2009. All rights reserved.                                                  //
// ========================================================================================== //

#region Using Statements
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
using HFShader.Tools;

#endregion

namespace HFShader.Camera
{


    #region CamMode class
    /// <summary>
    /// The class CamMode encapsulates the different camera modes
    /// </summary>
    public static class Generic_CamMode {
        public const int firstPerson = 0;
        public const int thirdPerson = 1;
        public const int isometric = 2;
        public const int numModes = 3;
    }
    #endregion




    #region GenericGameCamera -> Main api class

    /// <summary>
    /// The class GameCamera represents a generic game camera. This class is the interface
    /// to the game
    /// </summary>
    public class GenericGameCamera : InterfaceCamera
    {

        // declarations ////////////////////////////////////////////////////////
        #region Properties 



        // old camera position (can be used for substepping)
        private Vector3 mOldCameraPos;
        public Vector3 oldCameraPos
        {
            get { return mOldCameraPos; }
        }
		


        /// <summary>
        /// The camera mode
        /// </summary>
        private int mMode;
        public int mode
        {
            get { return mMode; }
            set
            {   if (value < 0) {
                    mMode = 0;
                }
                else if (value >= Generic_CamMode.numModes)
                {
                    mMode = 0;
                }
                else
                {
                    mMode = value;
                }
                mController.mode = mMode;
            }
        }


        private Vector3 mUpVector;
        public Vector3 upVector
        {
            get { return mUpVector; }
            set { mUpVector = value;
            mUpVector.Normalize();  }
        }


        // physical properties
        private float mStiffness;
        public float stiffnes {
            get {
                return mStiffness;
            }
        }

        private bool mPhysicallyComputed;
        public bool physicallyComputed
        {
            get { return mPhysicallyComputed; }
            set { mPhysicallyComputed = value; }
        }
	

        // geometrical properties
        private BoundingSphere mBoundingSphere;
        public BoundingSphere boundingSphere {
            get { return mBoundingSphere; }
            set { mBoundingSphere = value;
                  CameraRadius = value.Radius;
                }
        }

        private float CameraRadius = 3.0f;
        public float cameraRadius
        {
            get
            {
                return CameraRadius;
            }
            set
            {
                CameraRadius = value;
            }
        }

        private IRay mRay;
        public IRay intersectionRay
        {
            get { return mRay; }
        }


        public Vector3 mOldPlayerPos;
        public GameTime mCurGameTime;



        #endregion


        #region Camera Mode Parameter
        // TPC

        private float mTPCDistance;
        public float TPCDistance
        {
            get { return mTPCDistance; }
            set { mTPCDistance = value; }
        }

        private float mTPCAngle;
        public float TPCAngle // angle between the orientation and the camera in up direction [Degree]
        {
            get { return mTPCAngle; }
            set { mTPCAngle = value; }
        }

	    
        // FPC
        private Vector3 mFPCcamOffset;
        public Vector3 FPCcamOffset
        {
            get { return mFPCcamOffset; }
            set { mFPCcamOffset = value; }
        }

        // IC
        private float mICDistance;
        public float ICDistance
        {
            get { return mICDistance; }
            set { mICDistance = value; }
        }

        private Vector3 mICLookDir;
        public Vector3 ICLookDir
        {
            get { return mICLookDir; }
            set { mICLookDir = value; }
        }

        #endregion


        #region Constants
        // Constants ////////////////////////////////////////////////////////////
        private const float Gravity = -0.98f;
        private const float CameraMass = 0.08f;
        private const float Friction = 0.4f;
        #endregion


        #region Fields

        // Private Fields ///////////////////////////////////////////////////////
        private CamConfiguration mDesiredConfiguration;
        public CamConfiguration desiredConfiguration {
            set { mDesiredConfiguration = value; }
        }

        private float mPitchAngularVelocity;
        private float mYawAngularVelocity;
        private Generic_CameraController mController;
        private Vector3 mPositionVelocity;

        

        // collision parameter
        private Vector3 mAverageCollisionPoint;
        private Vector3 mCamSubstepPosition;
        private bool mCollides;

        // raycast parameter
        private float mIntersectT;
        #endregion


        // Functions ///////////////////////////////////////////////////////////
        #region Private Functions

        // applies a linear friction to an input vector
        private void applyLinearFriction(ref Vector3 input)
        {
            // linear friction
            input += Friction * -input;
        }


        // integrates the physical camera model
        private void integrate(Generic_Forces forces, GameTime gameTime)
        {
            // timestep
            float h = (float)gameTime.ElapsedGameTime.TotalMilliseconds / 1000.0f;


            // integrate position
            Vector3 acceleration = forces.positionForces / CameraMass;

            mPositionVelocity += Vector3.Multiply(acceleration, h);
            applyLinearFriction(ref mPositionVelocity);

            mConfiguration.position += Vector3.Multiply(mPositionVelocity, h);

            // integrate pitch
            double pitchAcc = forces.pitchForce / CameraMass;
            mPitchAngularVelocity += (float)(pitchAcc * h);
            mPitchAngularVelocity -= Friction * mPitchAngularVelocity;

            mConfiguration.pitch += mPitchAngularVelocity * h;

            // integrate yaw
            double yawAcc = forces.yawForce / CameraMass;
            mYawAngularVelocity += (float)(yawAcc * h);
            mYawAngularVelocity -= Friction * mYawAngularVelocity;

            mConfiguration.yaw += mYawAngularVelocity * h;
        }



        // resolves collisions if setCollisionParameter was called
        private Generic_Forces resolveCollisions(GameTime gameTime, Generic_Forces forces)
        {
            if (mCollides)
            {
                // average inverse collision direction
                Vector3 invCollisionVec = mCamSubstepPosition - mAverageCollisionPoint;
                float penDepth = invCollisionVec.Length();

                // normalized inv. collision direction
                Vector3 r = Vector3.Normalize(invCollisionVec);

                // reset camera position
                //mConfiguration.position = mAverageCollisionPoint + Vector3.Multiply(r, CameraRadius);

                // add penalty force
                float penaltyForceFactor = 500.0f;
                forces.positionForces += Vector3.Multiply(r, (CameraRadius - penDepth) / CameraRadius * penaltyForceFactor);

                mCollides = false;
            }
            return forces;
        }


        #endregion

        ///<summary>
        /// constructor
        ///</summary>
        public GenericGameCamera()
        {
            mUpVector = Vector3.Up;

            initialize(640.0f / 480.0f, 45.0f);

            mMode = Generic_CamMode.thirdPerson;
        }


        /// <summary>
        /// Initializes the generig game camera
        /// </summary>
        /// <param name="aspectRatio">The aspectratio of the camera (width/height)</param>
        /// <param name="openingAngle">The opening angle of the camera in degree</param>
        public void initialize(float aspectRatio, float openingAngle)
        {
            // init controller
            mController = new Generic_CameraController(this);
            mController.mode = Generic_CamMode.thirdPerson;


            // init physical properites
            mPositionVelocity = Vector3.Zero;
            mPitchAngularVelocity = 0.0f;
            mYawAngularVelocity  = 0.0f;


            // init camera settings
            mConfiguration = new CamConfiguration();
            mDesiredConfiguration = new CamConfiguration();
            mBoundingSphere = new BoundingSphere(mConfiguration.position, CameraRadius);

            mRay = new IRay();
            mRay.Position = Vector3.Zero;
            mRay.Direction = mUpVector;
            mRay.EndPosition = Vector3.Zero;
            mRay.Max_t = 0.0f;

            mStiffness = 2.5f;
            mPhysicallyComputed = true;

            mAspectRatio = aspectRatio;
            mOpeningAngle = openingAngle;
            mProjectionMatrix = Matrix.CreatePerspectiveFieldOfView(MathHelper.ToRadians(mOpeningAngle),
                                                                    mAspectRatio, mNearFarValues.X, mNearFarValues.Y);




            // create modelview matrix
            Matrix rollMatrix = Matrix.CreateRotationX(MathHelper.ToRadians(mConfiguration.roll));
            Matrix pitchMatrix = Matrix.CreateRotationY(MathHelper.ToRadians(mConfiguration.pitch));
            Matrix yawMatrix = Matrix.CreateRotationZ(MathHelper.ToRadians(mConfiguration.yaw));

            Vector3 up = Vector3.Transform(mUpVector, rollMatrix * yawMatrix * pitchMatrix);

            Matrix transformation = rollMatrix * yawMatrix * pitchMatrix;

            Vector3 lookDir = new Vector3(1.0f, 0.0f, 0.0f);
            lookDir = Vector3.Transform(lookDir, transformation);
            lookDir.Normalize();
            mViewMatrix = Matrix.CreateLookAt(mConfiguration.position, mConfiguration.position + lookDir, up);


            // init configuration
            mConfiguration = new CamConfiguration();


            // collision parameter
            mAverageCollisionPoint = Vector3.Zero;
            mCamSubstepPosition = Vector3.Zero;
            mCollides = false;
            mOldCameraPos = Vector3.Zero;

            mOldPlayerPos = Vector3.Zero;

            // raycast parameter
            mIntersectT = -1.0f;

            // set default parameter for camera modes
            mTPCAngle = 20.0f;
            mTPCDistance = 25.0f;

            mFPCcamOffset = new Vector3(0.0f, 1.5f, 0.0f);

            mICDistance = 60;
            mICLookDir = new Vector3(-1.0f, -1.0f, -1.0f);
        }

        

        /// <summary>
        /// when this function is called the camera handles in its next update() call the detected collision
        /// defined by the average collision point and the camera substep position
        /// </summary>
        /// <param name="avgCollisionPoint">The average point where the camera bounding sphere penetrates the geometry</param>
        /// <param name="camSubstepPosition">The position of the camera when penetrating the geometry</param>
        public void setCollisionParameter(Vector3 avgCollisionPoint, Vector3 camSubstepPosition)
        {
            mAverageCollisionPoint = avgCollisionPoint;
            mCamSubstepPosition = camSubstepPosition;
            mCollides = true;
        }


        /// <summary>
        /// This function handles a detected collision of the camera collision ray with the geometry. If t<0
        /// no collision is assumed.
        /// </summary>
        /// <param name="t">The parameter on the ray where the nearest collision to the camera collision ray's startpoint is detected</param>
        public void setRayCastParameter(float t)
        {
            mIntersectT = t;
        }



        /// <summary>
        /// updates the camera model
        /// </summary>
        /// <param name="gameTime">actual came time</param>
        /// <param name="playerPosition">position of the player</param>
        /// <param name="playerOrientation">the look direction of the player</param>
        /// <param name="gamePadState">player input</param>
        public void update(GameTime gameTime, Vector3 playerPosition, Vector3 playerOrientation)
        {

          

            // set old position 
            mOldCameraPos = mConfiguration.position;

            // set game time
            mCurGameTime = gameTime;


            // compute changes implied by the previous call of update using user input
            //////////////////////////////////////////////////////////////////////////
            
            // fix mDesiredConfiguration (wich was computed at the end of the previous update call) with ray collision parameter
            mDesiredConfiguration = mController.fixConfigurationWithRayIntersection(mDesiredConfiguration, mIntersectT);

            // move camera towards fixed configuration
            if (mPhysicallyComputed)
            {
                // move the camera towards the position where it should be
                Generic_Forces forces = mController.getForces(mConfiguration, mDesiredConfiguration, playerPosition, playerOrientation, mStiffness);
                resolveCollisions(gameTime, forces);
                integrate(forces, gameTime);
            }
            else
            {
                // just set the camera where it should be
                mConfiguration = mDesiredConfiguration;
            }
            


            // compute settings for the next update call (when user input is available)
            ////////////////////////////////////////////////////////////////

            // compute next desired configuration
            mDesiredConfiguration = mController.getDesiredConfiguration(mConfiguration, playerPosition, playerOrientation);

            // update bounding sphere
            mBoundingSphere = new BoundingSphere(mConfiguration.position, CameraRadius);

            // get new camera intersection ray that needs to be tested by the user before next update
            mRay = mController.getCollisionRay(playerPosition, playerOrientation);

            // setOldPlayerPos
            mOldPlayerPos = playerPosition;

            // reset ray intersection parameter
            mIntersectT = -1.0f;



            mViewMatrix = mConfiguration.createModelViewMatrix();

        }


    }


    #endregion




    #region controll routines


    class Generic_ControlRoutine
    {
        protected GenericGameCamera mCamera;

        protected IRay mCollisionRay;

        protected CamConfiguration mDesiredConfiguration;


        public Generic_ControlRoutine(ref GenericGameCamera camera)
        {
            mCamera = camera;
            mCollisionRay = new IRay();
            mDesiredConfiguration = new CamConfiguration();
        }

        public virtual CamConfiguration getDesiredConfiguration(CamConfiguration currentConfiguration, Vector3 playerPosition, Vector3 playerOrientation)
        {
            return new CamConfiguration();
        }

        public virtual IRay getRayToBeCollided(Vector3 playerPosition, Vector3 playerOrientation)
        {
            return mCollisionRay;
        }

        public virtual CamConfiguration fixConfWithRayCollision(CamConfiguration configuration, float intersectT)
        {
            return configuration;
        }

        public virtual Generic_Forces postprocessForces(Generic_Forces forces, CamConfiguration currentConfiguration, Vector3 playerPosition, Vector3 playerOrientation)
        {
            return forces;
        }
    }



    class Generic_FirstPersonCR : Generic_ControlRoutine
    {
        public Generic_FirstPersonCR(ref GenericGameCamera camera)
            : base(ref camera)
        {
        }

        public override CamConfiguration getDesiredConfiguration(CamConfiguration currentConfiguration, Vector3 playerPosition, Vector3 playerOrientation)
        {
            CamConfiguration configuration = new CamConfiguration();

            // desired position = player position 
            configuration.position = playerPosition + mCamera.FPCcamOffset;

            // compute desired pitch
            Vector3 orientXZ = new Vector3(playerOrientation.X, 0.0f, playerOrientation.Z);
            orientXZ.Normalize();

            float pitchAngle = (float)Math.Acos(Vector3.Dot(orientXZ, new Vector3(1.0f, 0.0f, 0.0f)));
            if (orientXZ.Z >= 0) // right
            {
                pitchAngle = 2 * (float)Math.PI - pitchAngle;
            }
            configuration.pitch = MathHelper.ToDegrees(pitchAngle);


            // compute desired yaw
            Vector3 orient = playerOrientation;
            orient.Normalize();
            float baseLen = (float)Math.Sqrt(orient.X * orient.X + orient.Z * orient.Z);
            float yawAngle = 0.0f;
            if (orient.Y > 0.001f)
            {
                yawAngle = (float)Math.Asin(orient.Y / 1.0);
                yawAngle = MathHelper.ToDegrees(yawAngle);
            }
            else if (orient.Y < -0.001f)
            {
                yawAngle = -(float)Math.Asin(Math.Abs(orient.Y) / 1.0);
                yawAngle = MathHelper.ToDegrees(yawAngle);
            }
            configuration.yaw = yawAngle;

            return configuration;
        }

        public override Generic_Forces postprocessForces(Generic_Forces forces, CamConfiguration currentConfiguration, Vector3 playerPosition, Vector3 playerOrientation)
        {
            Vector3 pos = playerPosition + mCamera.FPCcamOffset;
            Vector3 l = (currentConfiguration.position - pos);
            if (l.Length() < (mCamera.cameraRadius * 2.0f))
            {
                // lock camera to player position if fpc mode is active
                Vector3 delta = currentConfiguration.position - pos;
                currentConfiguration.position = pos - Vector3.Multiply(delta, 0.3f);

                // correct forces
                forces.positionForces = Vector3.Zero;
            }
            else
            {
                // not in range for locking... enhancing forces
                forces.positionForces = Vector3.Multiply(forces.positionForces, 3.0f);
            }
            return forces;
        }
    }



    class Generic_ThirdPersonCR : Generic_ControlRoutine
    {
        public Generic_ThirdPersonCR(ref GenericGameCamera camera)
            : base(ref camera)
        {
        }

        public override CamConfiguration getDesiredConfiguration(CamConfiguration currentConfiguration, Vector3 playerPosition, Vector3 playerOrientation)
        {
            CamConfiguration configuration = new CamConfiguration();

            // compute desired pitch

            // pitch oriented to desired orientation
            Vector3 orientXZ = new Vector3(playerOrientation.X, 0.0f, playerOrientation.Z);
            orientXZ.Normalize();

            float pitchAngle = (float)Math.Acos(Vector3.Dot(orientXZ, new Vector3(1.0f, 0.0f, 0.0f)));
            if (orientXZ.Z >= 0) // right
            {
                pitchAngle = 2 * (float)Math.PI - pitchAngle;
            }
            configuration.pitch = MathHelper.ToDegrees(pitchAngle);


            // pitch oriented towards player position
            //Vector3 orientPlayerXZ = playerPosition - currentConfiguration.position;
            //orientPlayerXZ.Y = 0.0f;
            //orientPlayerXZ.Normalize();
            //float pitchAnglePlayer = (float)Math.Acos(Vector3.Dot(orientXZ, new Vector3(1.0f, 0.0f, 0.0f)));
            //if (orientPlayerXZ.Z >= 0) // right
            //{
            //    pitchAnglePlayer = 2 * (float)Math.PI - pitchAnglePlayer;
            //}
            //configuration.pitch = MathHelper.ToDegrees(pitchAnglePlayer);


            // compute desired yaw
            configuration.yaw = -mCamera.TPCAngle;


            // compute offset direction
            Vector3 offset = new Vector3(-1.0f, (float)Math.Tan(MathHelper.ToRadians(mCamera.TPCAngle)), 0.0f);
            Matrix rot = Matrix.CreateRotationY(MathHelper.ToRadians(configuration.pitch));
            offset = Vector3.Transform(offset, rot);
            offset.Normalize();

            // compute desired position
            configuration.position = playerPosition + Vector3.Multiply(offset, mCamera.TPCDistance);

            // set configuration here to a global field because it will be used in later steps of the process
            mDesiredConfiguration = configuration;

            return configuration;
        }




        public override CamConfiguration fixConfWithRayCollision(CamConfiguration configuration, float intersectT)
        {
            // compute distance of cam and player (player position == mCollisionRay.position)
            Vector3 dist = mCollisionRay.Position - configuration.position;
            float l = dist.Length();

            // see if we have a ray intersection
            if ((intersectT > 0.0f) && (intersectT < l))
            {
                // ray intersection: fix the position
                //float t = intersectT - mCamera.cameraRadius;
                float t = intersectT - mCamera.cameraRadius;
                if (t < 0) t = 0.001f;

                // fix the configuration position
                configuration.position = mCollisionRay.Position + Vector3.Multiply(mCollisionRay.Direction, t - mCamera.cameraRadius);
            }
            return configuration;
        }




        public override Generic_Forces postprocessForces(Generic_Forces forces, CamConfiguration currentConfiguration, Vector3 playerPosition, Vector3 playerOrientation)
        {
            return forces;
        }

        public override IRay getRayToBeCollided(Vector3 playerPosition, Vector3 playerOrientation)
        {
            IRay ray = new IRay();
            ray.Position = playerPosition;
            ray.EndPosition = mDesiredConfiguration.position;
            ray.Direction = ray.EndPosition - ray.Position;
            ray.Max_t = ray.Direction.Length() + mCamera.cameraRadius;
            ray.Direction = Vector3.Normalize(ray.Direction);

            // set ray to a global field because it will be used later
            mCollisionRay = ray;

            return ray;
        }

    }


    class Generic_IsometricCR : Generic_ControlRoutine
    {
        public Generic_IsometricCR(ref GenericGameCamera camera)
            : base(ref camera)
        {
        }


        public override CamConfiguration getDesiredConfiguration(CamConfiguration currentConfiguration, Vector3 playerPosition, Vector3 playerOrientation)
        {
            CamConfiguration configuration = new CamConfiguration();

            // desired position = player position + offset
            Vector3 dir = mCamera.ICLookDir;
            dir.Normalize();
            configuration.position = playerPosition + Vector3.Multiply(dir, -mCamera.ICDistance);


            // compute desired pitch
            Vector3 orientXZ = new Vector3(mCamera.ICLookDir.X, 0.0f, mCamera.ICLookDir.Z);
            orientXZ.Normalize();

            float pitchAngle = (float)Math.Acos(Vector3.Dot(orientXZ, new Vector3(1.0f, 0.0f, 0.0f)));
            if (orientXZ.Z >= 0) // right
            {
                pitchAngle = 2 * (float)Math.PI - pitchAngle;
            }
            configuration.pitch = MathHelper.ToDegrees(pitchAngle);


            // compute desired yaw
            Vector3 orient = mCamera.ICLookDir;
            orient.Normalize();

            float yawAngle = (float)Math.Acos(Vector3.Dot(orientXZ, orient));
            if (orient.Z <= 0) // right
            {
                yawAngle = 2 * (float)Math.PI - yawAngle;
            }
            configuration.yaw = MathHelper.ToDegrees(yawAngle);

            // set global configuration field for later use
            mDesiredConfiguration = configuration;

            return configuration;
        }


        public override CamConfiguration fixConfWithRayCollision(CamConfiguration configuration, float intersectT)
        {
            //if ((intersectT > 0) && (intersectT < mCollisionRay.Max_t))
            //{
            //    // distance to ground to small... need to be set higher
            //    float delta_h = mCollisionRay.Max_t - intersectT;
            //    configuration.position += Vector3.Multiply(mCollisionRay.Direction, delta_h);

            //}
            return configuration;
        }


        public override IRay getRayToBeCollided(Vector3 playerPosition, Vector3 playerOrientation)
        {
            IRay ray = new IRay();
            ray.Position = playerPosition;
            ray.EndPosition = mDesiredConfiguration.position;
            ray.Direction = ray.EndPosition - ray.Position;
            ray.Max_t = ray.Direction.Length() + mCamera.cameraRadius;
            ray.Direction = Vector3.Normalize(ray.Direction);

            // set ray to a global field because it will be used later
            mCollisionRay = ray;

            return ray;
        }
    }


    #endregion




    #region

    class Generic_CameraController
    {


        private Generic_Servo mServo;

        private Generic_ControlRoutine[] mControllRoutines;

        private GenericGameCamera mCamera;


        private Vector3 mOldOrientation;
        private float mPitchDir;


        private int mMode;
        public int mode
        {
            get { return mMode; }
            set { mMode = value; }
        }



        public Generic_CameraController(GenericGameCamera camera)
        {
            mCamera = camera;

            mServo = new Generic_Servo();
            mControllRoutines = new Generic_ControlRoutine[Generic_CamMode.numModes];

            // init camera routines
            mControllRoutines[Generic_CamMode.firstPerson] = new Generic_FirstPersonCR(ref mCamera);
            mControllRoutines[Generic_CamMode.thirdPerson] = new Generic_ThirdPersonCR(ref mCamera);
            mControllRoutines[Generic_CamMode.isometric] = new Generic_IsometricCR(ref mCamera);

            // history parameter
            mOldOrientation = Vector3.Zero;
        }

        public IRay getCollisionRay(Vector3 playerPosition, Vector3 playerOrientation)
        {
            return mControllRoutines[mMode].getRayToBeCollided(playerPosition, playerOrientation);
        }

        public CamConfiguration getDesiredConfiguration(CamConfiguration currentConfiguration, Vector3 playerPosition, Vector3 playerOrientation)
        {
            return mControllRoutines[mMode].getDesiredConfiguration(currentConfiguration, playerPosition, playerOrientation);
        }

        public CamConfiguration fixConfigurationWithRayIntersection(CamConfiguration configuration, float intersectT)
        {
            return mControllRoutines[mMode].fixConfWithRayCollision(configuration, intersectT);
        }

        public Generic_Forces getForces(CamConfiguration currentConfiguration, CamConfiguration desiredConfiguration, Vector3 playerPosition, Vector3 playerOrientation, float stiffness)
        {

            // compute change in yaw direction
            Vector3 delta = playerOrientation - mOldOrientation;
            if (delta.Length() > 0.001f)
            { // orientation changed 
                delta.Normalize();
                float oldPitchDir = mPitchDir;
                mPitchDir = Math.Sign((Vector3.Cross(playerOrientation, delta)).Y);
                mOldOrientation = playerOrientation;
            }


            // get forces from servo for desired configuration
            mServo.stiffness = stiffness;
            Generic_Forces forces = mServo.getCameraForces(currentConfiguration, desiredConfiguration, mPitchDir);

            return mControllRoutines[mMode].postprocessForces(forces, mCamera.configuration, playerPosition, playerOrientation);
        }

    }


    #endregion




    #region forces & servo

    /// <summary>
    /// The class Forces encapsulates forces for the camera model
    /// </summary>
    class Generic_Forces
    {
        // forces applied to a position
        private Vector3 mPosForces;
        public Vector3 positionForces
        {
            get { return mPosForces; }
            set { mPosForces = value; }
        }

        // force applied to the pitch angle direction
        private double mPitchForce;
        public double pitchForce
        {
            get { return mPitchForce; }
            set { mPitchForce = value; }
        }

        // force applied to the yaw angle direction
        private double mYawForce;
        public double yawForce
        {
            get { return mYawForce; }
            set { mYawForce = value; }
        }


        public Generic_Forces()
        {
            mPosForces = Vector3.Zero;
            mPitchForce = 0.0f;
            mYawForce = 0.0f;
        }
    }


    class Generic_Servo
    {


        private float mStiffness;
        public float stiffness
        {
            set
            {
                mStiffness = MathHelper.Clamp(value, 0.0f, 5.0f);
            }
            get
            {
                return mStiffness;
            }
        }


        public Generic_Servo()
        {
            mStiffness = 1.0f;
        }




        public Generic_Forces getCameraForces(CamConfiguration currentConf, CamConfiguration desiredConf, float pitchDir)
        {
            // parameter
            //////////////////////////////
            float distanceFactor = 15.0f * mStiffness;
            float angularFactor = 20.0f * mStiffness;

            Generic_Forces forces = new Generic_Forces();


            // calculations
            /////////////////////////////

            // positional forces
            forces.positionForces = Vector3.Multiply(desiredConf.position - currentConf.position, distanceFactor);


            // pitch force
            forces.pitchForce = 0.0f;
            float diff = 0.0f;
            if (pitchDir > 0)
            {
                if (desiredConf.pitch < currentConf.pitch)
                {
                    diff = (desiredConf.pitch + 360.0f) - currentConf.pitch;
                }
                else
                {
                    diff = desiredConf.pitch - currentConf.pitch;
                }
            }
            else // pitchDir < 0 
            {
                if (desiredConf.pitch > currentConf.pitch)
                {
                    diff = (desiredConf.pitch - 360.0f) - currentConf.pitch;
                }
                else
                {
                    diff = desiredConf.pitch - currentConf.pitch;
                }
            }


            //if (Math.Sign(diff) != Math.Sign(pitchDir)) {
            //    diff = diff;
            //}
            if (diff > 180.0f) diff -= 360.0f;
            if (diff < -180.0f) diff += 360.0f;

            forces.pitchForce = diff * angularFactor;





            // yaw force
            diff = desiredConf.yaw - currentConf.yaw;
            if (diff > 180.0f)
            {
                diff -= 360.0f;
            }
            forces.yawForce = diff * angularFactor;

            return forces;
        }

    }



    #endregion



}
