// ========================================================================================== //
//             ___  _   _    __    __  __    __    ____  __    ____  _____  _  _              //
//            / __)( )_( )  /__\  (  \/  )  /__\  ( ___)(  )  ( ___)(  _  )( \( )             //
//           ( (__  ) _ (  /(__)\  )    (  /(__)\  )__)  )(__  )__)  )(_)(  )  (              //
//            \___)(_) (_)(__)(__)(_/\/\_)(__)(__)(____)(____)(____)(_____)(_)\_)             //
//            -------------------------------------------- G A M E   E N G I N E              //
//                                                                                            //
// ========================================================================================== //
// Developed and implemented by:                                                              //
//     Mader Theodor                                                                          //
//     Oskam Thomas                                                                           //
//                                                                                            //
// Copyright 2007-2009. All rights reserved.                                                  //
// ========================================================================================== //

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using HFShader.Camera;
using HFShader.Tools;


namespace HFShader.Rendering
{

    /// <summary>
    /// This class defines different render modes the heightfield class offers.
    /// </summary>
    public class HeightfieldRenderMode {
        public const int UseGivenNormalTexture = 0;
        public const int ComputeNormals = 1;
        public const int PrecomputeNormalTexture = 2;
    }


    /// <summary>
    /// This class contains the set of textures that a heightfield needs to render. Also the preloaded 
    /// heightfield shader is passed through this object (for now, needs to be outsourced to the renderer maybe).
    /// </summary>
    public class HeightfieldParameter 
    {
        // Dimensional parameter
        public float HeightfieldWorldWidth;
        public float HeightfieldWorldHeight;

        public Vector3 Translation;

        // Light Parameter
        public Vector3 LightDirection;


        // blending Parameter
        public Vector2 Alpha_Start_End;

        // Textures
        public Texture2D HeightfieldTexture;
        public Texture2D ColorChannelTexture;
        public Texture2D NormalTexture;

        public string RedChannelTexture;
        public string GreenChannelTexture;
        public string BlueChannelTexture;
        public string AlphaChannelTexture;

        // local texture stretch factors
        public float RedStretchFactor;
        public float GreenStretchFactor;
        public float BlueStretchFactor;
        public float AlphaStretchFactor;

        public int RenderMode;
    }




    /// <summary>
    /// This is a helper class that holds information of the bounding volume, index, and center of 
    /// a heightfield patch.
    /// </summary>
    public class HFPatchBox {
        public BoundingBox boundingBox;

        public Vector2 patchIndex;

        public Vector3 patchCenter;

        public int lodLevel;
    }



    /// <summary>
    /// This class defines a scene graph (SG) node. It is the basic structure from which all objects, that
    /// want to be able to be included in the scene tree (and thus rendered), need to inherit. SGNode inherits
    /// from MessageCapable, and thus, provides post man functionality
    /// </summary>
    public class Heightfield
    {



        // for threat safety
        private Object mLocker = new Object();


        #region constants
        const int SIZE_OF_VECTOR2 = sizeof(float) * 2;
        const int DEF_PATCH_SIZE = 32;
        #endregion




        // defined heightfield parameter
        HeightfieldParameter mParameter;
        float[] mHeightmapBuffer;

        Effect mHeightfieldShader;
        Effect mZPassShader;
        Effect mShadowedHFShader;


        Texture2D mRedChannelText;
        Texture2D mGreenChannelText;
        Texture2D mBlueChannelText;
        Texture2D mAlphaChannelText;

        // parameters
        private int mTextureSizeX;
        private int mTextureSizeY;
        private int mNOfMeshTris;
        private int mNOfMeshVerts;


        private float mPatchWorldSize;
        private int mPatchSize; // grid dimension of one patch
        private int mPatchDimensionX; // number of patches to cover heightfield, in X dimension
        private int mPatchDimensionY; // number of patches to cover heightfield, in Y dimension
        private float mPixelWorldDelta; // width of a pixel in world coordinates
        private Vector3 mPixelWorldOffset; // offset in world coords to move the field in order to allign the nodes with pixel centers


        // buffers
        private VertexBuffer mMeshBuffer;
        private IndexBuffer mMeshIndices;
        private Vector2[] mPatchIndexArray;
        private Vector2[] mFullPatchIndexArray;

        VertexBuffer mInstanceDataStream;

        // height data buffer
        //private float[] mHeightTextureData;

        private GraphicsDevice mGraphicsDevice;

        private VertexDeclaration mInstancedMeshVertexDeclaration;

        // shader parameter pointers
        EffectParameter SHADER_viewProjection;
        EffectParameter SHADER_heightTexture;
        EffectParameter SHADER_lightDirection;

        EffectParameter Z_SHADER_heightTexture;
        EffectParameter Z_SHADER_viewProjection;


        // map data structure (for culling lod etc)
        public HFPatchBox[,] mPatchBoxMatrix;


        // content manager
        ContentManager mContentManager;

        // properties
        #region props

        private Vector3 mLightDirection;
        public Vector3 lightDirection
        {
            get { return mLightDirection; }
            set { mLightDirection = value; }
        }


        #endregion



        /// <summary>
        /// Heightfield constructor. Sets parameters, initializes grid-patch, prepares shader
        /// </summary> 
        public Heightfield(HeightfieldParameter hfp, ContentManager contentManager, GraphicsDevice graphicsDevice, string name)
        {
            // store parameter objects
            mContentManager = contentManager;

            mParameter = hfp;
            mHeightmapBuffer = new float[hfp.HeightfieldTexture.Width * hfp.HeightfieldTexture.Height];
            lock (mLocker)
            {
                mParameter.HeightfieldTexture.GetData<float>(mHeightmapBuffer);
            }


            // load channel textures
            mRedChannelText = contentManager.Load<Texture2D>(hfp.RedChannelTexture);
            mGreenChannelText = contentManager.Load<Texture2D>(hfp.GreenChannelTexture);
            mBlueChannelText = contentManager.Load<Texture2D>(hfp.BlueChannelTexture);
            mAlphaChannelText = contentManager.Load<Texture2D>(hfp.AlphaChannelTexture);


            // compute further parameter
            #region compute further parameter

            // copy light direction to property field (dynamic lighting anyone?)
            mLightDirection = hfp.LightDirection;

            // set grid
            mTextureSizeX = mParameter.HeightfieldTexture.Width;
            mTextureSizeY = mParameter.HeightfieldTexture.Height;

            // get reference to graphics device
            mGraphicsDevice = graphicsDevice;

            // define patch size (x and y dimension are the same)
            mPatchSize = DEF_PATCH_SIZE;

            // compute the number of patches in X and Y dimension 
            mPatchDimensionX = mTextureSizeX / (mPatchSize);
            mPatchDimensionY = mTextureSizeY / (mPatchSize);

            // compute the worldsiye of a patch
            mPatchWorldSize = mParameter.HeightfieldWorldWidth / mPatchDimensionX;

            // compute the size of a heightmap pixel in world size
            mPixelWorldDelta = mParameter.HeightfieldWorldWidth / (float)mTextureSizeX;

            // compute the 1/2 pixel width offset of the heightfield mesh in world size
            mPixelWorldOffset = new Vector3(mPixelWorldDelta * 0.5f, 0.0f, mPixelWorldDelta * 0.5f);

            // build the patch vertex buffer (highest level)
            buildVertexAndIndexBuffer(mPatchSize, mPatchWorldSize, mPatchDimensionX, mPatchDimensionY);

            #endregion

            // load & init zPass shader
            #region load & init zPass shader
            mZPassShader = contentManager.Load<Effect>("Content/Shader/heightfieldZPassShader");
            mZPassShader.CurrentTechnique = mZPassShader.Techniques["fastZPass"];

            // set heightfield
            mZPassShader.Parameters["xHeightfield"].SetValue(mParameter.HeightfieldTexture);
            mZPassShader.Parameters["xMaxAlteringHeight"].SetValue(mParameter.HeightfieldWorldHeight);
            mZPassShader.Parameters["xPatchWidth"].SetValue(mPatchWorldSize);
            mZPassShader.Parameters["xHeightfieldPixelOffset"].SetValue(new Vector2(1.0f / (float)mPatchDimensionX, 1.0f / (float)mPatchDimensionY));
            mZPassShader.Parameters["xHeightfieldTranslation"].SetValue(mPixelWorldOffset + mParameter.Translation);
            #endregion


            // load & init zPass shader
            #region load & init shadowed hf shader
            mShadowedHFShader = contentManager.Load<Effect>("Content/Shader/shadowedHeightfieldShader");
            mShadowedHFShader.CurrentTechnique = mShadowedHFShader.Techniques["shadowedHeightfield"];

            // set heightfield
            mShadowedHFShader.Parameters["xHeightfield"].SetValue(mParameter.HeightfieldTexture);
            mShadowedHFShader.Parameters["xMaxAlteringHeight"].SetValue(mParameter.HeightfieldWorldHeight);
            mShadowedHFShader.Parameters["xPatchWidth"].SetValue(mPatchWorldSize);
            mShadowedHFShader.Parameters["xHeightfieldPixelOffset"].SetValue(new Vector2(1.0f / (float)mPatchDimensionX, 1.0f / (float)mPatchDimensionY));
            mShadowedHFShader.Parameters["xHeightfieldTranslation"].SetValue(mPixelWorldOffset + mParameter.Translation);
            #endregion

            // load & init heightfield shader
            #region load shader and set defualt global parameter
            mHeightfieldShader = contentManager.Load<Effect>("Content/Shader/heightfieldShader");

            // init shader technique given the specified render mode
            switch (mParameter.RenderMode)
            {
                case HeightfieldRenderMode.UseGivenNormalTexture:
                    mHeightfieldShader.CurrentTechnique = mHeightfieldShader.Techniques["instancedHeightfielded"];
                    mHeightfieldShader.Parameters["xNormalTexture"].SetValue(mParameter.NormalTexture);
                    break;
                case HeightfieldRenderMode.ComputeNormals:
                    mHeightfieldShader.CurrentTechnique = mHeightfieldShader.Techniques["instancedHeightfieldedComputeNormals"];
                    break;
                case HeightfieldRenderMode.PrecomputeNormalTexture:
                    mHeightfieldShader.CurrentTechnique = mHeightfieldShader.Techniques["instancedHeightfielded"];
                    Texture2D newNormalTexture = computeNormalTexture(mParameter.HeightfieldTexture);
                    mHeightfieldShader.Parameters["xNormalTexture"].SetValue(newNormalTexture);
                    // save normal texture
                    newNormalTexture.Save("newNormalTexture.dds", ImageFileFormat.Dds);
                    newNormalTexture.Save("newNormalTexture.tga", ImageFileFormat.Tga);
                    break;
            }
            // set textures
            mHeightfieldShader.Parameters["xHeightfield"].SetValue(mParameter.HeightfieldTexture);
            mHeightfieldShader.Parameters["xChannelfield"].SetValue(mParameter.ColorChannelTexture);

            mHeightfieldShader.Parameters["xRedChannelTexture"].SetValue(mRedChannelText);
            mHeightfieldShader.Parameters["xGreenChannelTexture"].SetValue(mGreenChannelText);
            mHeightfieldShader.Parameters["xBlueChannelTexture"].SetValue(mBlueChannelText);
            mHeightfieldShader.Parameters["xAlphaChannelTexture"].SetValue(mAlphaChannelText);
            
            // set other shader parameter
            mHeightfieldShader.Parameters["localTextureStretchFactor"].SetValue(new Vector4(hfp.RedStretchFactor, hfp.GreenStretchFactor, hfp.BlueStretchFactor, hfp.AlphaStretchFactor));
            mHeightfieldShader.Parameters["xTexCoordDelta"].SetValue(new Vector2(1.0f / (float)mTextureSizeX, 1.0f / (float)mTextureSizeY));
            mHeightfieldShader.Parameters["xTexCoordWorldDelta"].SetValue(mPixelWorldDelta);
            mHeightfieldShader.Parameters["xMaxAlteringHeight"].SetValue(mParameter.HeightfieldWorldHeight);
            mHeightfieldShader.Parameters["xPatchWidth"].SetValue(mPatchWorldSize);
            mHeightfieldShader.Parameters["xHeightfieldPixelOffset"].SetValue(new Vector2(1.0f / (float)mPatchDimensionX, 1.0f / (float)mPatchDimensionY));
            mHeightfieldShader.Parameters["xHeightfieldTranslation"].SetValue(mPixelWorldOffset + mParameter.Translation);
            mHeightfieldShader.Parameters["xAlpha_Start_End"].SetValue(hfp.Alpha_Start_End);

            mHeightfieldShader.Parameters["localYTextureRelation"].SetValue((float)mPatchDimensionY / (float)mPatchDimensionX);
            

            #endregion


            // build patch box data structure
            mPatchBoxMatrix = buildPatchBoxMatrix(mPatchSize);

            // create shader parameter pointers
            createShaderParameterPointer();

            // init instance data stream
            int instanceDataSize = SIZE_OF_VECTOR2 * mPatchDimensionX * mPatchDimensionY;
            mInstanceDataStream = new VertexBuffer(mGraphicsDevice, instanceDataSize, BufferUsage.None);
            mPatchBoxMatrix = buildPatchBoxMatrix(mPatchSize);
        }



        #region PREPROCESSING



        public void prepareShadowedHFShader(int RST_ShadowMap_Resolution, uint baseSplits)
        {
            int linBaseSplits = 0;
            for (int i = 0; i < baseSplits; i++)
            {
                linBaseSplits += (i + 1);
            }

            // load shadowmap shader
            mShadowedHFShader.Parameters["xMaxDepth"].SetValue(1700.0f);
            mShadowedHFShader.Parameters["xMinDistanceLight"].SetValue(1200.0f);
            mShadowedHFShader.Parameters["xMapResolution"].SetValue(RST_ShadowMap_Resolution);
            mShadowedHFShader.Parameters["xLogBaseSplitNum"].SetValue(linBaseSplits);
        }

        public void setShadowedSceneDrawParameter(ref RenderTarget2D[] mShadowMapRenderTarget, ref Matrix[] mShadowProjections, Vector3 mGlobalLightDirection, Vector2[] mFrustumSplitNearFars, InterfaceCamera camera)
        {
            int num_shadow_maps = mShadowMapRenderTarget.Length;
            for (int i = 0; i < num_shadow_maps; i++)
            {
                mShadowedHFShader.Parameters["xShadowMap" + i.ToString()].SetValue(mShadowMapRenderTarget[i].GetTexture());
                mShadowedHFShader.Parameters["xLightViewProjection" + i.ToString()].SetValue(mShadowProjections[i]);
            }
            mShadowedHFShader.Parameters["xNumShadowMaps"].SetValue(num_shadow_maps);

            // set global shader parameter
            mShadowedHFShader.Parameters["xLightDirection"].SetValue(mGlobalLightDirection);
            mShadowedHFShader.Parameters["xView"].SetValue(camera.viewMatrix);
            mShadowedHFShader.Parameters["xProjection"].SetValue(camera.projectionMatrix);

            // get camera near-far plane values
            Vector2 camNearFar = camera.nearFarValues;

            //camera.nearFarValues = new Vector2(mFrustumSplitNearFars[0].X, mFrustumSplitNearFars[0].Y);
            float farRange = (mFrustumSplitNearFars[num_shadow_maps - 1].Y - mFrustumSplitNearFars[0].X) / (camNearFar.Y - camNearFar.X);
            mShadowedHFShader.Parameters["xNearFarRange"].SetValue(new Vector2(mFrustumSplitNearFars[0].X, mFrustumSplitNearFars[num_shadow_maps - 1].Y));
            mShadowedHFShader.Parameters["xCamNearFar"].SetValue(camera.nearFarValues);
        }




        /// <summary>
        /// Creates pointers to specific shader parameter that change during runtime. These pointers are 
        /// globally stored for faster parameter change at runtime.
        /// </summary>
        private void createShaderParameterPointer() {
            // Pass camera matrices through to the effect.
            SHADER_viewProjection = mHeightfieldShader.Parameters["xViewProjection"];
            //SHADER_cameraPosition = mHeightfieldShader.Parameters["xCameraPosition"];
            SHADER_heightTexture = mHeightfieldShader.Parameters["xHeightfield"];
            SHADER_lightDirection = mHeightfieldShader.Parameters["xLightDirection"];

            Z_SHADER_heightTexture = mZPassShader.Parameters["xHeightfield"];
            Z_SHADER_viewProjection = mZPassShader.Parameters["xViewProjection"];
        }


        /// <summary>
        /// Builds the vertex and index buffer of a mesh patch with the passed size (cells of the mesh). Also
        /// defines the global parameter of the mesh such as mNOfMeshTris, mNOfMeshVerts and mInstancedMeshVertexDeclaration.
        /// </summary>
        /// <param name="gridSize">Number of cells of the grid in one dimension. A grid patch is always quadratic.</param>
        /// <param name="gridWorldLen">The length of the grid patch in world coordinates</param>
        /// <param name="patchDim">The number of grid patches per dimension to cover the whole heightmap.</param>
        private void buildVertexAndIndexBuffer(int gridSize, float gridWorldLen, int patchesDimX, int patchesDimY)
        {

            // 0) Init Parameter 
            /////////////////////////////////////////////////////////////////////////////////////

            int S = gridSize;                 // grid-size 
            float dL = gridWorldLen / S;      // cell-width

            float texCoordFactorX = 1.0f / (float)patchesDimX;
            float texCoordFactorY = 1.0f / (float)patchesDimY;


            mNOfMeshVerts = (S + 1) * (S + 1);


            
            // 1) Build Vertex Buffer for patch mesh
            ////////////////////////////////////////////////////////////////////////////////////
            
            // define basic vertex format
            VertexDeclaration basicVertexDeclaration = new VertexDeclaration(mGraphicsDevice, 
                                                                             VertexPositionTexture.VertexElements);

            // create instancing vertex format
            mInstancedMeshVertexDeclaration = getInstancedVertexDeclaration(basicVertexDeclaration);


            // create vertex buffer
            VertexBuffer vertexBuffer = new VertexBuffer(mGraphicsDevice, 
                                                         VertexPositionTexture.SizeInBytes * mNOfMeshVerts, 
                                                         BufferUsage.None);

            // create data array
            VertexPositionTexture[] dataArray = new VertexPositionTexture[mNOfMeshVerts];


            


            // (x,z) is a vertex in the SxS grid (with S+1 vertices per dimension)
            int dataIndex = 0;
            for (int z = 0; z < S+1; z++)
            {
                for (int x = 0; x < S+1; x++)
                {
                    // Compute Vertex for (x,z)
                    // vertex position
                    Vector3 pos = new Vector3(x * dL, 0.0f, z * dL);
                    Vector2 texCoord = new Vector2((float)x / (float)S, ((float)z / (float)S));
                    VertexPositionTexture v = new VertexPositionTexture(pos, texCoord * new Vector2(texCoordFactorX, texCoordFactorY));
                    dataArray[dataIndex] = v;

                    dataIndex++;
                }
            }


            // set data array to vertex buffer
            vertexBuffer.SetData<VertexPositionTexture>(dataArray);

            // store vertex buffer
            mMeshBuffer = vertexBuffer;


                // 3) create index buffer
                /////////////////////////////////////////////////////////////////


                mNOfMeshTris = S * S * 2; // two triangles per mesh cell

            // init index list
            short[] gridIndices = new short[mNOfMeshTris * 3];

            // create tris
            // by looping through all mesh cells (i,j) element [(0,0), (S-1, S-1)]

            // using the following grid structure for a 4x4 texture
            //                                          v1
            //    |---|---|---|                         |
            //    |\5 |\  |\17|           triangle1:    |\       
            //    |4\ | \ | \ |                         | \ 
            //    |--\|--\|--\|                      v3 |--\  v2
            //    |\3 |\  |\  |
            //    |2\ | \ | \ |
            //    |--\|--\|--\|           triangle2: v4 |---| v5
            //   |\1 |\7 |\  |                          \  | 
            // |  |0\ |6\ | \ |                           \ |  
            // x  |--\|--\|--\|                            \| 
            //    z -->                                     v6
            //
            // Vertices of the grid lie on the texture pixels

            int indexIndex = 0;
            for (int j = 0; j < S; j++)
            {
                for (int i = 0; i < S; i++)
                {
                    // first triangle
                    gridIndices[indexIndex + 0] = (short)((S + 1)*(j + 1) + i);
                    gridIndices[indexIndex + 1] = (short)((S + 1) * j + i);
                    gridIndices[indexIndex + 2] = (short)((S + 1)* j      + i + 1);

                    // second triangle
                    gridIndices[indexIndex + 3] = (short)((S + 1) * (j + 1) + i);
                    gridIndices[indexIndex + 4] = (short)((S + 1) * j + i + 1);
                    gridIndices[indexIndex + 5] = (short)((S + 1) * (j + 1) + i + 1);


                    indexIndex += 6;
                }
            }



            mMeshIndices = new IndexBuffer(mGraphicsDevice, sizeof(short) * mNOfMeshTris * 3, BufferUsage.None, IndexElementSize.SixteenBits);
            mMeshIndices.SetData<short>(gridIndices);



            // 3) create patch index array for instancing
            ////////////////////////////////////////////////////////////////
            // build index vectors (the list of the patches that need to be instanced) [here all... later we can cull away some]
            mFullPatchIndexArray = new Vector2[mPatchDimensionX * mPatchDimensionY];
            for (int y = 0; y < mPatchDimensionY; y++)
            {
                for (int x = 0; x < mPatchDimensionX; x++)
                {
                    mFullPatchIndexArray[x + mPatchDimensionX * y] = new Vector2(x, y);
                }
            }
        }



        #region computation of normal texture given a heightmap


        private Texture2D computeNormalTexture(Texture2D heightfield) {
            int textW = heightfield.Width;
            int textH = heightfield.Height;

            // create new texture
            // TODO: see that this texture gets killed when the content loader is unloaded
            Texture2D normalTexture = new Texture2D(mGraphicsDevice, 
                                                    textW, 
                                                    textH, 1, 
                                                    TextureUsage.None, 
                                                    SurfaceFormat.Vector4);

            // define normalTexture buffer
            Vector4[] rgba = new Vector4[textW * textH];


            // define parameters
            float two_texDelta = mPixelWorldDelta + mPixelWorldDelta;
            float divisor = 1.0f / two_texDelta;

            // get values from heightfield textue
            float[] heightmapBuffer = new float[textW * textH];
            // mGraphicsDevice.Textures[0] = null;
            lock (mLocker)
            {
                mParameter.HeightfieldTexture.GetData<float>(mHeightmapBuffer);
            }

            for (int y = 0; y < textH; y++) {
                for (int x = 0; x < textW; x++) {

                    // collect indices of required neighbors
                    Vector2 northClamped = clampCoords(x, y + 1, textW, textH);
                    Vector2 southClamped = clampCoords(x, y - 1, textW, textH);
                    Vector2 eastClamped = clampCoords(x + 1, y, textW, textH);
                    Vector2 westClamped = clampCoords(x - 1, y, textW, textH);
                    int centerIndex = get1DIndex(x, y, textW);
                    int northIndex = get1DIndex((int)northClamped.X, (int)northClamped.Y, textW);
                    int southIndex = get1DIndex((int)southClamped.X, (int)southClamped.Y, textW);
                    int eastIndex = get1DIndex((int)eastClamped.X, (int)eastClamped.Y, textW);
                    int westIndex = get1DIndex((int)westClamped.X, (int)westClamped.Y, textW);


                    // collect height values
                    float center_h = heightmapBuffer[centerIndex];
                    float north_h = heightmapBuffer[northIndex];
                    float south_h = heightmapBuffer[southIndex];
                    float east_h = heightmapBuffer[eastIndex];
                    float west_h = heightmapBuffer[westIndex];

                    
                    // compute normal 
                    Vector3 south_minus_north = new Vector3( 0.0f, (south_h - north_h) * mParameter.HeightfieldWorldHeight, -(two_texDelta) );
                    Vector3 east_minus_west = new Vector3( -(two_texDelta), (east_h - west_h) * mParameter.HeightfieldWorldHeight, 0.0f );
                    Vector3 normal = Vector3.Normalize(Vector3.Cross((south_minus_north) * divisor, (east_minus_west) * divisor));
                
                    // compute byte values of the vectors
                    Vector4 c = new Vector4(normal.X, normal.Y, normal.Z, 1.0f);

                    rgba[centerIndex] = c;

                }
            }


            normalTexture.SetData<Vector4>(rgba);
            return normalTexture;        
        }

        #region helper funcs for normal texture computation
        private Vector2 clampCoords(int coordX, int coordY, int textSizeX, int textSizeY) {
            return new Vector2( MathHelper.Clamp(coordX, 0, textSizeX - 1), MathHelper.Clamp(coordY, 0, textSizeY -1 ) );
        }


        private int get1DIndex(int i, int j, int i_dim)
        {
            return i + j * i_dim;
        }
        #endregion

        #endregion



        private HFPatchBox[,] buildPatchBoxMatrix(int patchSize) { 

            // get texture size
            int mapWidth = mParameter.HeightfieldTexture.Width;
            int mapHeight = mParameter.HeightfieldTexture.Height;

            // compute patch matrix size
            int matrixWidth = (int)(mapWidth / patchSize);
            int matrixHeight = (int)(mapHeight / patchSize);

            // grab heightmap texture data
            float[] heightmapData = new float[mapWidth * mapHeight];
            // mGraphicsDevice.Textures[0] = null;
            lock (mLocker)
            {
                mParameter.HeightfieldTexture.GetData<float>(mHeightmapBuffer);
            }

            //  alloc memory
            HFPatchBox[,] patchBoxMatrix = new HFPatchBox[matrixWidth, matrixHeight];


            // compute patch bounding boxes and patchIndices 
            for (int y = 0; y < matrixHeight; y++) {
                for (int x = 0; x < matrixWidth; x++) {

                    HFPatchBox newPatchBox = new HFPatchBox();

                    // set patchIndex
                    newPatchBox.patchIndex = new Vector2(x, y);


                    // set bb values (first x and z only)
                    Vector3 translationXZ = new Vector3(mParameter.Translation.X, 0.0f, mParameter.Translation.Z);
                    translationXZ += mPixelWorldOffset;
                    float v2 = 0.0f;
                    v2 += mPixelWorldDelta;
                    newPatchBox.boundingBox = new BoundingBox(); 
                    newPatchBox.boundingBox.Min = new Vector3(x * mPatchWorldSize, -100.0f, y * mPatchWorldSize) + translationXZ;
                    newPatchBox.boundingBox.Max = new Vector3((x + 1) * mPatchWorldSize, 100.0f, (y + 1) * mPatchWorldSize) + translationXZ;


                    // compute min and max height values
                    float localMapMax = -float.MaxValue;
                    float localMapMin = float.MaxValue;
                    #region go through heightmapData values the patch covers

                    int iStart = x * mPatchSize;
                    int jStart = y * mPatchSize;

                    for (int i = iStart; i < iStart + patchSize; i++)
                    {
                        for (int j = jStart; j < jStart + patchSize; j++)
                        {
                            float height = heightmapData[get1DIndex(i, j, mapWidth)];
                            if (height > localMapMax) localMapMax = height;
                            if (height < localMapMin) localMapMin = height;
                        }
                    }
                    #endregion

                    float h_min = localMapMin * mParameter.HeightfieldWorldHeight + mParameter.Translation.Y;
                    float h_max = localMapMax * mParameter.HeightfieldWorldHeight + mParameter.Translation.Y;

                    // set min and max height values
                    newPatchBox.boundingBox.Min.Y = h_min;
                    newPatchBox.boundingBox.Max.Y = h_max;

                    // set default lod
                    newPatchBox.lodLevel = 0;

                    // compute patch center
                    newPatchBox.patchCenter = Vector3.Multiply((newPatchBox.boundingBox.Min + newPatchBox.boundingBox.Max), 0.5f);

                    // insert patch into matrix
                    patchBoxMatrix[x, y] = newPatchBox;
                }
            }




            return patchBoxMatrix;
        } 


        #endregion



        #region RENDERING



        public void draw(InterfaceCamera camera)
        {

            #region View Frustum & distance culling precomputations (old code; not used)
            //// compute camera frustum
            //BoundingFrustum camFrustum = new BoundingFrustum(camera.viewMatrix * camera.projectionMatrix);
            //float halvPatchWidth = 0.5f * mPatchWorldSize;
            //float savetyDelta = 0.5f * mPatchWorldSize;

            //// create new patchArray buffer (distance culling)
            //List<Vector2> visiblePatches = new List<Vector2>();
            //for (int i = 0; i < mFullPatchIndexArray.Length; i++)
            //{
            //    Vector2 curPatch = mFullPatchIndexArray[i];

            //    Vector3 patchPos = new Vector3(curPatch.X * mPatchWorldSize + halvPatchWidth, 0.0f, curPatch.Y * mPatchWorldSize + halvPatchWidth) + mParameter.Translation;
            //    float d = Vector3.Distance(camera.configuration.position, patchPos);
            //    if ((d - 1.5 * mPatchWorldSize) < mParameter.Alpha_Start_End.Y * 0.8)
            //    {
            //        BoundingBox patchBB = new BoundingBox();

            //        // TODO: compute height for each patch
            //        Vector3 bbOffset = new Vector3(halvPatchWidth + savetyDelta, 300.0f, halvPatchWidth + savetyDelta);

            //        patchBB.Min = patchPos - bbOffset;
            //        patchBB.Max = patchPos + bbOffset;

            //        if (patchBB.Intersects(camFrustum)) visiblePatches.Add(curPatch);
            //    }
            //}
            //// build array from list
            //mPatchIndexArray = new Vector2[visiblePatches.Count];
            //visiblePatches.CopyTo(mPatchIndexArray);
            #endregion

            // create new patch index list
            List<Vector2> visiblePatches = new List<Vector2>();
            
            BoundingFrustum camFrustum = new BoundingFrustum(camera.viewMatrix * camera.projectionMatrix);
            for (int y = 0; y < mPatchDimensionY; y++) {
                for (int x = 0; x < mPatchDimensionX; x++)
                {
                    // 1) test against heightfield
                    //if ((mPatchBoxMatrix[x, y].boundingBox.Intersects(camFrustum)))
                        // 2) test against distance
                        //if (Vector3.Distance(mPatchBoxMatrix[x, y].patchCenter, camera.configuration.position) < (mParameter.Alpha_Start_End.Y + mPatchWorldSize * one_over_sqrt2))
                        //if (Vector3.Distance(mPatchBoxMatrix[x, y].patchCenter, camera.configuration.position) < 600.0f)
                            // ==> add patch to list
                            visiblePatches.Add(new Vector2(x, y));
                }
            }
            
            // copy list to patch index buffer
            mPatchIndexArray = new Vector2[visiblePatches.Count];
            visiblePatches.CopyTo(mPatchIndexArray);

            

            //mPatchIndexArray = mFullPatchIndexArray;




            // set alpha blending
            mGraphicsDevice.RenderState.AlphaBlendEnable = true;
            mGraphicsDevice.RenderState.SourceBlend = Blend.SourceAlpha;
            mGraphicsDevice.RenderState.DestinationBlend = Blend.InverseSourceAlpha;

            // draw z-pass
            zPassDraw(camera);

            // draw heightfield
            instancedDraw(camera);

            // disable alpha
            mGraphicsDevice.RenderState.AlphaBlendEnable = false;
        }



        /// <summary>
        /// Draws the heightfield using shader 3.0 hardware instancing.
        /// </summary>
        /// <param name="camera">The camera for which the heightfield is drawn</param>
        private void instancedDraw(InterfaceCamera camera)
        {

            // security check if patch buffer is null
            if (mPatchIndexArray.Length == 0) return;

            // Pass camera matrices through to the effect.
            SHADER_viewProjection.SetValue(camera.viewMatrix * camera.projectionMatrix);
            SHADER_lightDirection.SetValue(mLightDirection);
            mHeightfieldShader.Parameters["xCameraPosition"].SetValue(Vector3.Zero);

            // set vertex buffer stream
            // Upload transform matrices to the instance data vertex buffer.
            //mInstanceDataStream.SetData(mPatchIndexArray, 0, mPatchIndexArray.Length);

            // start shader
            // set vertex declaration
            //mGraphicsDevice.VertexDeclaration = mInstancedMeshVertexDeclaration;


            // Set up two vertex streams for instanced rendering.
            // The first stream provides the actual vertex data, while
            // the second provides per-instance transform matrices.
            //mGraphicsDevice.Vertices[0].SetSource(mMeshBuffer, 0, VertexPositionTexture.SizeInBytes);
            //mGraphicsDevice.Indices = mMeshIndices;
            //mGraphicsDevice.Vertices[0].SetFrequencyOfIndexData(mPatchIndexArray.Length);
            //mGraphicsDevice.Vertices[1].SetSource(mInstanceDataStream, 0, SIZE_OF_VECTOR2);
            //mGraphicsDevice.Vertices[1].SetFrequencyOfInstanceData(1);

            
            mHeightfieldShader.Begin();
            mHeightfieldShader.CurrentTechnique.Passes[0].Begin();


            // call draw
            mGraphicsDevice.DrawIndexedPrimitives(
                PrimitiveType.TriangleList,
                0,
                0,
                mNOfMeshTris * 3,
                0,
                mNOfMeshTris
            );



            // end technique
            mHeightfieldShader.CurrentTechnique.Passes[0].End();
            mHeightfieldShader.End();


            // Reset the instancing streams.
            mGraphicsDevice.Vertices[0].SetSource(null, 0, 0);
            mGraphicsDevice.Vertices[1].SetSource(null, 0, 0);
        }


        /// <summary>
        /// Draws the heightfield using shader 3.0 hardware instancing.
        /// </summary>
        /// <param name="camera">The camera for which the heightfield is drawn</param>
        private void zPassDraw(InterfaceCamera camera)
        {
            // security check if patch buffer is null
            if (mPatchIndexArray.Length == 0) return;


            // Pass camera matrices through to the effect.
            Z_SHADER_viewProjection.SetValue(camera.viewMatrix * camera.projectionMatrix);

            //mZPassShader.CurrentTechnique = mZPassShader.Techniques[0];

            // set vertex buffer stream
            // Upload transform matrices to the instance data vertex buffer.
            mInstanceDataStream.SetData(mPatchIndexArray, 0, mPatchIndexArray.Length);

            // start shader
            // set vertex declaration
            mGraphicsDevice.VertexDeclaration = mInstancedMeshVertexDeclaration;


            // Set up two vertex streams for instanced rendering.
            // The first stream provides the actual vertex data, while
            // the second provides per-instance transform matrices.
            mGraphicsDevice.Vertices[0].SetSource(mMeshBuffer, 0, VertexPositionTexture.SizeInBytes);
            mGraphicsDevice.Indices = mMeshIndices;
            mGraphicsDevice.Vertices[0].SetFrequencyOfIndexData(mPatchIndexArray.Length);
            mGraphicsDevice.Vertices[1].SetSource(mInstanceDataStream, 0, SIZE_OF_VECTOR2);
            mGraphicsDevice.Vertices[1].SetFrequencyOfInstanceData(1);


            mZPassShader.Begin();
            mZPassShader.CurrentTechnique.Passes[0].Begin();

            // call draw
            mGraphicsDevice.DrawIndexedPrimitives(
                PrimitiveType.TriangleList,
                0,
                0,
                mNOfMeshTris * 3,
                0,
                mNOfMeshTris
            );

            // end technique
            mZPassShader.CurrentTechnique.Passes[0].End();
            mZPassShader.End();


            // Reset the instancing streams.
            //mGraphicsDevice.Vertices[0].SetSource(null, 0, 0);
            //mGraphicsDevice.Vertices[1].SetSource(null, 0, 0);
        }



        public void shadowedSceneDraw(InterfaceCamera camera) {

            // create new patch index list
            List<Vector2> visiblePatches = new List<Vector2>();

            Vector2[] m = new Vector2[mPatchDimensionX*mPatchDimensionY];            
            for (int y = 0; y < mPatchDimensionY; y++)
            {
                for (int x = 0; x < mPatchDimensionX; x++)
                {
                    m[mPatchDimensionY * x + y] = new Vector2(x,y);
                }
            }



            // security check if patch buffer is null
            if (m.Length == 0) return;


            // Pass camera matrices through to the effect.
            mShadowedHFShader.Parameters["xViewProjection"].SetValue(camera.viewMatrix * camera.projectionMatrix);

            // set vertex buffer stream
            // Upload transform matrices to the instance data vertex buffer.
            mInstanceDataStream.SetData(mPatchIndexArray, 0, m.Length);

            // start shader
            // set vertex declaration
            mGraphicsDevice.VertexDeclaration = mInstancedMeshVertexDeclaration;


            // Set up two vertex streams for instanced rendering.
            // The first stream provides the actual vertex data, while
            // the second provides per-instance transform matrices.
            mGraphicsDevice.Vertices[0].SetSource(mMeshBuffer, 0, VertexPositionTexture.SizeInBytes);
            mGraphicsDevice.Indices = mMeshIndices;
            mGraphicsDevice.Vertices[0].SetFrequencyOfIndexData(m.Length);
            mGraphicsDevice.Vertices[1].SetSource(mInstanceDataStream, 0, SIZE_OF_VECTOR2);
            mGraphicsDevice.Vertices[1].SetFrequencyOfInstanceData(1);


            mShadowedHFShader.Begin();
            mShadowedHFShader.CurrentTechnique.Passes[0].Begin();

            // call draw
            mGraphicsDevice.DrawIndexedPrimitives(
                PrimitiveType.TriangleList,
                0,
                0,
                mNOfMeshTris * 3,
                0,
                mNOfMeshTris
            );

            // end technique
            mShadowedHFShader.CurrentTechnique.Passes[0].End();
            mShadowedHFShader.End();

            mGraphicsDevice.Vertices[0].SetSource(null, 0, 0);
            mGraphicsDevice.Vertices[1].SetSource(null, 0, 0);
        }



        /// <summary>
        /// Builds an extended vertex declaration that can be used to pass the extra i,j coordinates
        /// to the shader in order to properly hardware instance the patches for the heightmap.
        /// </summary>
        /// <param name="standardDeclaration">Original vertex declaration</param>
        /// <returns>Extended vertex declaration</returns>
        private VertexDeclaration getInstancedVertexDeclaration(VertexDeclaration standardDeclaration)
        {
            VertexDeclaration result;

            VertexElement[] extraElements = new VertexElement[1];

            short offset = 0;
            byte usageIndex = 2;
            short stream = 1;


            extraElements[0] = new VertexElement(stream, offset,
                                                 VertexElementFormat.Vector2,
                                                 VertexElementMethod.Default,
                                                 VertexElementUsage.TextureCoordinate,
                                                 usageIndex
                                                 );


            // Append the new elements to the original format.
            int originalLength = standardDeclaration.GetVertexElements().Length;
            int length = originalLength + extraElements.Length;

            VertexElement[] elements = new VertexElement[length];
            standardDeclaration.GetVertexElements().CopyTo(elements, 0);
            extraElements.CopyTo(elements, originalLength);

            // Create a new vertex declaration.
            result = new VertexDeclaration(mGraphicsDevice, elements);

            return result;
        }










        #endregion




        #region COLLISION DETECTION

        // PUBLIC //////////////////////////////////////////////////////////////////////////////////////////////////////

        public List<ITriangle> collideAABB(BoundingBox aabb)
        {

            List<ITriangle> collidingTris = new List<ITriangle>();

            // HACK
            if (float.IsNaN(aabb.Min.X)) return collidingTris;
            // HACK

            // 1) compute the range of texture coordinates that are occupied by the aabb's x-z projection
            Vector2 min2 = new Vector2(aabb.Min.X, aabb.Min.Z);
            Vector2 max2 = new Vector2(aabb.Max.X, aabb.Max.Z);

            Vector2 trans2 = new Vector2(mParameter.Translation.X, mParameter.Translation.Z);

            // move min2 and max2 in x-z direction such that the coordinates are w.r.t the texture coordidinates
            min2 -= trans2;
            max2 -= trans2;

            // scale the position down to [0,textureSize] of text coords
            float divisorX = (float)(mTextureSizeX) / mParameter.HeightfieldWorldWidth;

            Vector2 delta = new Vector2(0.5f, 0.5f);
            min2 = Vector2.Multiply(min2, divisorX) - delta;
            max2 = Vector2.Multiply(max2, divisorX) + delta;

            // check if aabb is inside heightfield
            if ((min2.X > (float)mTextureSizeX) || (min2.Y > (float)mTextureSizeY) || (max2.X < 0.0f) || (max2.Y < 0.0f))
            {
                return collidingTris;
            }

            // okay now lets find tex coord range
            min2 = new Vector2(MathHelper.Max((int)Math.Floor(min2.X - 0.5f), 0), MathHelper.Max((int)Math.Floor(min2.Y - 0.5f), 0));
            max2 = new Vector2(MathHelper.Min((int)Math.Floor(max2.X + 0.5f), mTextureSizeX - 1), MathHelper.Min((int)Math.Floor(max2.Y + 0.5f), mTextureSizeY - 1));

            // 2) generate potential colliding triangles

            float pixelWorldDelta = mPixelWorldDelta;
            Vector3 pixelWorldOffset = mPixelWorldOffset + new Vector3(mParameter.Translation.X, mParameter.Translation.Y, mParameter.Translation.Z);


            // get height values
            int dx = (int)(max2.X - min2.X + 1);
            int dy = (int)(max2.Y - min2.Y + 1);

            float[] pixelColor = new float[dx * dy];

            Rectangle pixelPatch = new Rectangle((int)min2.X, (int)min2.Y, dx, dy);
            // // mGraphicsDevice.Textures[0] = null;
            lock (mLocker)
            {
                mParameter.HeightfieldTexture.GetData<float>(0, pixelPatch, pixelColor, 0, dx * dy);
            }

            // generate triangles
            for (int y = (int)min2.Y; y < (int)max2.Y; y++)
            {
                int pixel_y = y - (int)min2.Y;
                for (int x = (int)min2.X; x < (int)max2.X; x++)
                {
                    int pixel_x = x - (int)min2.X;

                    Vector3 pixelWorldPos = pixelWorldOffset + new Vector3(x * pixelWorldDelta, mParameter.HeightfieldWorldHeight * pixelColor[pixel_x + pixel_y * dx], y * pixelWorldDelta);

                    Vector3 pixelEastWorldPos = pixelWorldOffset + new Vector3((x + 1) * pixelWorldDelta, mParameter.HeightfieldWorldHeight * pixelColor[(pixel_x + 1) + pixel_y * dx], y * pixelWorldDelta);

                    Vector3 pixelNorthWorldPos = pixelWorldOffset + new Vector3(x * pixelWorldDelta, mParameter.HeightfieldWorldHeight * pixelColor[pixel_x + (pixel_y + 1) * dx], (y + 1) * pixelWorldDelta);

                    Vector3 pixelNorthEastWorldPos = pixelWorldOffset + new Vector3((x + 1) * pixelWorldDelta, mParameter.HeightfieldWorldHeight * pixelColor[(pixel_x + 1) + (pixel_y + 1) * dx], (y + 1) * pixelWorldDelta);

                    ITriangle t1 = new ITriangle(pixelWorldPos, pixelEastWorldPos, pixelNorthWorldPos);
                    ITriangle t2 = new ITriangle(pixelEastWorldPos, pixelNorthEastWorldPos, pixelNorthWorldPos);
                    t1.computeNormal();
                    t2.computeNormal();

                    if (triangleInsideAABBHeight(t1, aabb.Min.Y, aabb.Max.Y)) collidingTris.Add(t1);
                    if (triangleInsideAABBHeight(t1, aabb.Min.Y, aabb.Max.Y)) collidingTris.Add(t2);

                }
            }
            return collidingTris;
        }

        public List<ITriangle> collideSphere(BoundingSphere sphere, ref List<Vector3> resultingNearestPoints)
        { 
            Vector3 radiusVec = Vector3.Multiply(Vector3.One, sphere.Radius);
            BoundingBox sphereAABB = new BoundingBox(sphere.Center - radiusVec, sphere.Center+radiusVec);
            return collideTrianglesSphere(sphere, collideAABB(sphereAABB), ref resultingNearestPoints);
        }

        public float intersectRay(IRay ray, ref ITriangle collidingTriangle) {

            // 1) compute aabb
            Vector3 deltaBox = new Vector3(4.0f, 0.0f, 4.0f);
            Vector3 min = new Vector3(MathHelper.Min(ray.Position.X, ray.EndPosition.X), MathHelper.Min(ray.Position.Y, ray.EndPosition.Y), MathHelper.Min(ray.Position.Z, ray.EndPosition.Z));
            Vector3 max = new Vector3(MathHelper.Max(ray.Position.X, ray.EndPosition.X), MathHelper.Max(ray.Position.Y, ray.EndPosition.Y), MathHelper.Max(ray.Position.Z, ray.EndPosition.Z));
            BoundingBox aabb = new BoundingBox(min - deltaBox, max + deltaBox);

            // get triangles in aabb
            List<ITriangle> potentialTriangles = collideAABB(aabb);
            
            // intersect with triangles
            float min_t = ray.Max_t + 1.0f;
            for (int i = 0; i < potentialTriangles.Count; i++) {
                ITriangle tri = potentialTriangles[i];

                float t = GeometryHelper.intersectRayTriangle(ray, tri.P0, tri.P1, tri.P2);
                if (t >= 0) {
                    if (t < min_t) {
                        collidingTriangle = tri;
                        min_t = t;                    
                    }
                }

            }

            if (min_t > ray.Max_t)
            {
                return -1;
            }
            else return min_t;
        }


        public bool findNearestPixelToRayIntersect(IRay ray, ref Vector2 nearestPixel, ref Vector3 pixel3Dpos)
        {
            // 0) compute aabb
            Vector3 deltaBox = new Vector3(4.0f, 0.0f, 4.0f);
            Vector3 min = new Vector3(MathHelper.Min(ray.Position.X, ray.EndPosition.X), MathHelper.Min(ray.Position.Y, ray.EndPosition.Y), MathHelper.Min(ray.Position.Z, ray.EndPosition.Z));
            Vector3 max = new Vector3(MathHelper.Max(ray.Position.X, ray.EndPosition.X), MathHelper.Max(ray.Position.Y, ray.EndPosition.Y), MathHelper.Max(ray.Position.Z, ray.EndPosition.Z));
            BoundingBox aabb = new BoundingBox(min - deltaBox, max + deltaBox);

            // 1) compute the range of texture coordinates that are occupied by the aabb's x-z projection
            Vector2 min2 = new Vector2(aabb.Min.X, aabb.Min.Z);
            Vector2 max2 = new Vector2(aabb.Max.X, aabb.Max.Z);

            Vector2 trans2 = new Vector2(mParameter.Translation.X, mParameter.Translation.Z);

            // move min2 and max2 in x-z direction such that the coordinates are w.r.t the texture coordidinates
            min2 -= trans2;
            max2 -= trans2;

            // scale the position down to [0,textureSize] of text coords
            float divisorX = (float)(mTextureSizeX) / mParameter.HeightfieldWorldWidth;

            Vector2 delta = new Vector2(0.5f, 0.5f);
            min2 = Vector2.Multiply(min2, divisorX) - delta;
            max2 = Vector2.Multiply(max2, divisorX) + delta;

            // check if aabb is inside heightfield
            if ((min2.X > (float)mTextureSizeX) || (min2.Y > (float)mTextureSizeY) || (max2.X < 0.0f) || (max2.Y < 0.0f))
            {
                return false;
            }

            // okay now lets find tex coord range
            min2 = new Vector2(MathHelper.Max((int)Math.Floor(min2.X - 0.5f), 0), MathHelper.Max((int)Math.Floor(min2.Y - 0.5f), 0));
            max2 = new Vector2(MathHelper.Min((int)Math.Floor(max2.X + 0.5f), mTextureSizeX - 1), MathHelper.Min((int)Math.Floor(max2.Y + 0.5f), mTextureSizeY - 1));

            // 2) generate potential colliding triangles

            float pixelWorldDelta = mPixelWorldDelta;
            Vector3 pixelWorldOffset = mPixelWorldOffset + new Vector3(mParameter.Translation.X, mParameter.Translation.Y, mParameter.Translation.Z);


            // get height values
            int dx = (int)(max2.X - min2.X + 1);
            int dy = (int)(max2.Y - min2.Y + 1);

            float[] pixelColor = new float[dx * dy];

            Rectangle pixelPatch = new Rectangle((int)min2.X, (int)min2.Y, dx, dy);
            // // mGraphicsDevice.Textures[0] = null;
            lock (mLocker)
            {
                mParameter.HeightfieldTexture.GetData<float>(0, pixelPatch, pixelColor, 0, dx * dy);
            }
            // prepare variables
            float nearestDistance = float.MaxValue;
            bool found = false;

            // go through all and test for nearest
            for (int y = (int)min2.Y; y <= (int)max2.Y; y++)
            {
                int pixel_y = y - (int)min2.Y;
                for (int x = (int)min2.X; x <= (int)max2.X; x++)
                {
                    int pixel_x = x - (int)min2.X;

                    Vector3 pixelWorldPos = pixelWorldOffset + new Vector3(x * pixelWorldDelta, mParameter.HeightfieldWorldHeight * pixelColor[pixel_x + pixel_y * dx], y * pixelWorldDelta);

                    // compute distance of pixel to ray
                    float dist = Vector3.Distance(GeometryHelper.getNearestPointOnRay(ray, pixelWorldPos), pixelWorldPos);

                    // if better
                    if (dist < nearestDistance) {
                        nearestPixel = new Vector2(x, y);
                        nearestDistance = dist;
                        pixel3Dpos = pixelWorldPos;
                        found = true;
                    }

                }
            }

            // returns true if found
            return found;
        }


        public Vector2 getTextureCoords(Vector3 worldPosition) {


            // 1) compute the texccords of the x-z projection of worldPosition
            Vector2 pos2 = new Vector2(worldPosition.X, worldPosition.Z);
            Vector2 trans2 = new Vector2(mParameter.Translation.X, mParameter.Translation.Z);

            // apply translation
            pos2 -= trans2;

            // scale the position down to [0,textureSize] of text coords
            float divisor = (float)(mTextureSizeX - 1) / mParameter.HeightfieldWorldWidth;
            pos2 = Vector2.Multiply(pos2, divisor);

            // clamp to texture width
            pos2.X = (int)Math.Round(MathHelper.Clamp(pos2.X, 0, mTextureSizeX - 1));
            pos2.Y = (int)Math.Round(MathHelper.Clamp(pos2.Y, 0, mTextureSizeY - 1));

            // return result
            return pos2;
        }


        public List<ITexTriangle> generateHFMeshTriangles()
        {
            return generateGeometry(new Vector2(0,0), new Vector2(mTextureSizeX, mTextureSizeY));
        }

        // PRIVATE //////////////////////////////////////////////////////////////////////////////////////////////////////


        // generates triangles for the given texture coordinates
        private List<ITexTriangle> generateGeometry(Vector2 texMin, Vector2 texMax)
        {
            List<ITexTriangle> generatedTris = new List<ITexTriangle>();

            // 1) clamp texture values to (0, textureSize - 1)
            texMin.X = MathHelper.Max(texMin.X, 0);
            texMin.Y = MathHelper.Max(texMin.Y, 0);

            texMax.X = MathHelper.Min(texMax.X, mTextureSizeX - 1);
            texMax.Y = MathHelper.Min(texMax.Y, mTextureSizeY - 1);


            // 2) test if given coords are valid
            if (texMin.X >= texMax.X) return generatedTris;
            if (texMin.Y >= texMax.Y) return generatedTris;

            // 3) generate triangles
            int dx = (int)((int)texMax.X - (int)texMin.X + 1);
            int dy = (int)((int)texMax.Y - (int)texMin.Y + 1);

            float[] pixelColor = new float[(dx) * (dy)];

            Rectangle pixelPatch = new Rectangle((int)texMin.X, (int)texMin.Y, dx, dy);
            // mGraphicsDevice.Textures[0] = null;
            lock (mLocker)
            {
                mParameter.HeightfieldTexture.GetData<float>(0, pixelPatch, pixelColor, 0, (dx) * (dy));
            }

            // store offsets
            float pixelWorldDelta = mPixelWorldDelta;
            Vector3 pixelWorldOffset = mPixelWorldOffset + new Vector3(mParameter.Translation.X, mParameter.Translation.Y, mParameter.Translation.Z);

            // generate triangles
            for (int y = (int)texMin.Y; y < (int)texMax.Y; y++)
            {
                int pixel_y = y - (int)texMin.Y;
                for (int x = (int)texMin.X; x < (int)texMax.X; x++)
                {
                    int pixel_x = x - (int)texMin.X;

                    Vector3 pixelWorldPos = pixelWorldOffset + new Vector3(x * pixelWorldDelta, mParameter.HeightfieldWorldHeight * pixelColor[pixel_x + pixel_y * dx], y * pixelWorldDelta);
                    Vector2 texCpos = new Vector2((float)pixel_x / (float)(dx - 1), (float)pixel_y / (float)(dy - 1));

                    Vector3 pixelEastWorldPos = pixelWorldOffset + new Vector3((x + 1) * pixelWorldDelta, mParameter.HeightfieldWorldHeight * pixelColor[(pixel_x + 1) + pixel_y * dx], y * pixelWorldDelta);
                    Vector2 texCeast = new Vector2((float)(pixel_x + 1) / (float)(dx - 1), (float)pixel_y / (float)(dy - 1));

                    Vector3 pixelNorthWorldPos = pixelWorldOffset + new Vector3(x * pixelWorldDelta, mParameter.HeightfieldWorldHeight * pixelColor[pixel_x + (pixel_y + 1) * dx], (y + 1) * pixelWorldDelta);
                    Vector2 texCnorth = new Vector2((float)pixel_x / (float)(dx - 1), (float)(pixel_y + 1) / (float)(dy - 1));

                    Vector3 pixelNorthEastWorldPos = pixelWorldOffset + new Vector3((x + 1) * pixelWorldDelta, mParameter.HeightfieldWorldHeight * pixelColor[(pixel_x + 1) + (pixel_y + 1) * dx], (y + 1) * pixelWorldDelta);
                    Vector2 texCnortheast = new Vector2((float)(pixel_x + 1) / (float)(dx - 1), (float)(pixel_y + 1) / (float)(dy - 1));

                    ITexTriangle t1 = new ITexTriangle(pixelNorthWorldPos, pixelEastWorldPos, pixelWorldPos);
                    t1.setTexCoords(texCnorth, texCeast, texCpos);

                    ITexTriangle t2 = new ITexTriangle(pixelNorthWorldPos, pixelNorthEastWorldPos, pixelEastWorldPos);
                    t2.setTexCoords(texCnorth, texCnortheast, texCeast);


                    // add tris
                    generatedTris.Add(t1);
                    generatedTris.Add(t2);

                }
            }

            return generatedTris;
        }


        private bool triangleInsideAABBHeight(ITriangle t, float minH, float maxH)
        {
            float minT = MathHelper.Min(MathHelper.Min(t.P0.Y, t.P1.Y), t.P2.Y);
            if (minT > maxH) return false;

            float maxT = MathHelper.Max(MathHelper.Max(t.P0.Y, t.P1.Y), t.P2.Y);
            if (maxT < minH) return false;

            return true;
        }


        private List<ITriangle> collideTrianglesSphere(BoundingSphere sphere, List<ITriangle> triangles, ref List<Vector3> resultingNearestPoints)
        {
            List<ITriangle> collidingTris = new List<ITriangle>();
            resultingNearestPoints = new List<Vector3>();




            for (int i = 0; i < triangles.Count; i++) {
                ITriangle triangle = triangles[i];

                // get parameter
                Vector3 pos1 = triangle.P0;
                Vector3 pos2 = triangle.P1;
                Vector3 pos3 = triangle.P2;

                BoundingSphere b = sphere;

                float squaredBoundingSphereRadius = b.Radius * b.Radius;

                Vector3 closestPoint = GeometryHelper.closestPointInTriangle(b.Center, pos1, pos2, pos3);
                float squaredDist = (closestPoint - b.Center).LengthSquared();
                if (squaredDist <= squaredBoundingSphereRadius)
                {
                    collidingTris.Add(triangle);
                    resultingNearestPoints.Add(closestPoint);
                }


            }
            return collidingTris;
        }


        private bool pointInTriangle( Vector3 pt1, Vector3 pt2, Vector3 pt3, Vector3 ptTest ) {
            // Compute vectors        
            Vector3 v0 = pt3 - pt1;
            Vector3 v1 = pt2 - pt1;
            Vector3 v2 = ptTest - pt1;

            // Compute dot products
            float dot00 = Vector3.Dot( v0, v0 );
            float dot01 = Vector3.Dot(v0, v1);
            float dot02 = Vector3.Dot(v0, v2);
            float dot11 = Vector3.Dot(v1, v1);
            float dot12 = Vector3.Dot(v1, v2);

            // Compute barycentric coordinates
            float invDenom = 1 / (dot00 * dot11 - dot01 * dot01);
            float u = (dot11 * dot02 - dot01 * dot12) * invDenom;
            float v = (dot00 * dot12 - dot01 * dot02) * invDenom;

            if( u+v<=1 && u>0 && v>0 )
                return true;
            return false;
       }

            



        #endregion



        #region REAL TIME MANIPULATION

        public void replaceHeightfieldTexture(Texture2D newHeightTexture) {
            // dispose current texture
            mParameter.HeightfieldTexture.Dispose();

            // set parameter texture
            mParameter.HeightfieldTexture = newHeightTexture;

            // repull buffer
            mHeightmapBuffer = new float[mParameter.HeightfieldTexture.Width * mParameter.HeightfieldTexture.Height];
            // mGraphicsDevice.Textures[0] = null;
            lock (mLocker)
            {
                mParameter.HeightfieldTexture.GetData<float>(mHeightmapBuffer);
            }
            // loat texture to shaders
            SHADER_heightTexture.SetValue(newHeightTexture);
            Z_SHADER_heightTexture.SetValue(newHeightTexture);

            // recompute boundingboxes
            mPatchBoxMatrix = buildPatchBoxMatrix(mPatchSize);

        }


        public void replaceColorChannelTexture(Texture2D newColormap)
        {
            mParameter.ColorChannelTexture.Dispose();

            mParameter.ColorChannelTexture = newColormap;
            mHeightfieldShader.Parameters["xChannelfield"].SetValue(mParameter.ColorChannelTexture);
        }


        public void setHeightfieldWorldWidth(float worldWidth)
        {
            mParameter.HeightfieldWorldWidth = worldWidth;

            mPatchWorldSize = mParameter.HeightfieldWorldWidth / mPatchDimensionX;
            mHeightfieldShader.Parameters["xPatchWidth"].SetValue(mPatchWorldSize);
            mZPassShader.Parameters["xPatchWidth"].SetValue(mPatchWorldSize);

            mPixelWorldDelta = mParameter.HeightfieldWorldWidth / (float)mTextureSizeX;
            mHeightfieldShader.Parameters["xTexCoordWorldDelta"].SetValue(mPixelWorldDelta);

            // define
            float localTextStretch = 40.0f;
            Vector2 localTextureStretchFactor = new Vector2(mParameter.HeightfieldWorldWidth / localTextStretch, mParameter.HeightfieldWorldWidth * (mTextureSizeY / mTextureSizeX) / localTextStretch);

            // build the patch vertex buffer (highest level)
            buildVertexAndIndexBuffer(mPatchSize, mPatchWorldSize, mPatchDimensionX, mPatchDimensionY);
        }


        public void setHeightfieldWorldHeight(float worldHeight) {
            mParameter.HeightfieldWorldHeight = worldHeight;
            mHeightfieldShader.Parameters["xMaxAlteringHeight"].SetValue(mParameter.HeightfieldWorldHeight);
            mZPassShader.Parameters["xMaxAlteringHeight"].SetValue(mParameter.HeightfieldWorldHeight);
        }


        public void setAlphaParameter(Vector2 alphaStartEnd) {
            mParameter.Alpha_Start_End = alphaStartEnd;
            mHeightfieldShader.Parameters["xAlpha_Start_End"].SetValue(alphaStartEnd);
        }


        public void setTranslation(Vector3 translation) {
            mParameter.Translation = translation;
            mHeightfieldShader.Parameters["xHeightfieldTranslation"].SetValue(mPixelWorldOffset + mParameter.Translation);
            mZPassShader.Parameters["xHeightfieldTranslation"].SetValue(mPixelWorldOffset + mParameter.Translation);
        }


        public HeightfieldParameter getHeightfieldParameterSnapshot() {

            if (mParameter.RenderMode == HeightfieldRenderMode.ComputeNormals)
            {

                // compute normal texture
                mParameter.NormalTexture = computeNormalTexture(mParameter.HeightfieldTexture);
            }
        
            // return parameter
            return mParameter;
        }


        public void replaceRedChannelTexture(string newTexture) { 
            // try load texture
            try
            {
                // load texture
                Texture2D newText = mContentManager.Load<Texture2D>(newTexture);

                newText.GenerateMipMaps(TextureFilter.Anisotropic);

                // unload current texture
                //mRedChannelText.Dispose();

                // set new texture
                mRedChannelText = newText;

                // set parameter
                mParameter.RedChannelTexture = newTexture;

                // set texture to shader
                mHeightfieldShader.Parameters["xRedChannelTexture"].SetValue(mRedChannelText);
            }
            catch (Exception)
            {
                return;
            } 
        }

        public void replaceGreenChannelTexture(string newTexture)
        {
            // try load texture
            try
            {
                // load texture
                Texture2D newText = mContentManager.Load<Texture2D>(newTexture);

                newText.GenerateMipMaps(TextureFilter.Anisotropic);

                // unload current texture
                //mGreenChannelText.Dispose();

                // set parameter
                mParameter.GreenChannelTexture = newTexture;

                // set new texture
                mGreenChannelText = newText;

                // set texture to shader
                mHeightfieldShader.Parameters["xGreenChannelTexture"].SetValue(mGreenChannelText);
            }
            catch (Exception)
            {
                return;
            }
        }

        public void replaceBlueChannelTexture(string newTexture)
        {
            // try load texture
            try
            {
                // load texture
                Texture2D newText = mContentManager.Load<Texture2D>(newTexture);

                newText.GenerateMipMaps(TextureFilter.Anisotropic);

                // unload current texture
                //mBlueChannelText.Dispose();

                // set parameter
                mParameter.BlueChannelTexture = newTexture;

                // set new texture
                mBlueChannelText = newText;

                // set texture to shader
                mHeightfieldShader.Parameters["xBlueChannelTexture"].SetValue(mBlueChannelText);
            }
            catch (Exception)
            {
                return;
            }
        }

        public void replaceAlphaChannelTexture(string newTexture)
        {
            // try load texture
            try
            {
                // load texture
                Texture2D newText = mContentManager.Load<Texture2D>(newTexture);

                newText.GenerateMipMaps(TextureFilter.Anisotropic);

                // unload current texture
                //mAlphaChannelText.Dispose();

                // set parameter
                mParameter.AlphaChannelTexture = newTexture;

                // set new texture
                mAlphaChannelText = newText;

                // set texture to shader
                mHeightfieldShader.Parameters["xAlphaChannelTexture"].SetValue(mAlphaChannelText);
            }
            catch (Exception)
            {
                return;
            }
        }






        public void setLocalTextStretchFactors(Vector4 stretches) {
            mParameter.RedStretchFactor = stretches.X;
            mParameter.GreenStretchFactor = stretches.Y;
            mParameter.BlueStretchFactor = stretches.Z;
            mParameter.AlphaStretchFactor = stretches.W;
            mHeightfieldShader.Parameters["localTextureStretchFactor"].SetValue(stretches);            
        }





        #endregion



        #region PIXEL QUERIES

        public Vector3 get3DPosition(Vector2 pixelPosition) { 
            
            // compute index
            int x = (int)pixelPosition.X;
            int y = (int)pixelPosition.Y;
            int index = TextureHelper.get1Dindex(x, y, mParameter.HeightfieldTexture.Width);

            // get height
            float pixelColor = mHeightmapBuffer[index];

            // compute position
            Vector3 pixelWorldOffset = mPixelWorldOffset + mParameter.Translation;
            Vector3 worldPos = pixelWorldOffset + new Vector3(x * mPixelWorldDelta, mParameter.HeightfieldWorldHeight * pixelColor, y * mPixelWorldDelta);
        
            // return result
            return worldPos;
        }


        public List<Vector3> get3DPosition(List<Vector2> pixelPositions) {
            List<Vector3> worldPositions = new List<Vector3>();
            for (int i = 0; i < pixelPositions.Count; i++) worldPositions.Add(get3DPosition(pixelPositions[i]));
            return worldPositions;
        }

        public List<Vector3> get3DPosition(List<Vector2> pixelPositions, Vector2 offsetPixel)
        {
            List<Vector3> worldPositions = new List<Vector3>();
            for (int i = 0; i < pixelPositions.Count; i++) worldPositions.Add(get3DPosition(pixelPositions[i] + offsetPixel));
            return worldPositions;
        }



        public bool isPixelInside(Vector2 pixelPosition) {
            int x = (int)pixelPosition.X;
            int y = (int)pixelPosition.Y;

            // return false if outside
            if (x < 0) return false;
            if (x >= mParameter.HeightfieldTexture.Width) return false;
            if (y < 0) return false;
            if (y >= mParameter.HeightfieldTexture.Height) return false;

            return true;
        }


        public List<Vector2> removeOutsidePixel(List<Vector2> pixelPositions) {
            List<Vector2> checkedPositions = new List<Vector2>();

            for (int i = 0; i < pixelPositions.Count; i++) {
                if (isPixelInside(pixelPositions[i])) checkedPositions.Add(pixelPositions[i]);
            }

            return checkedPositions;        
        }


        public List<Vector2> removeOutsidePixel(List<Vector2> pixelPositions, Vector2 offsetPixel)
        {
            List<Vector2> checkedPositions = new List<Vector2>();

            for (int i = 0; i < pixelPositions.Count; i++)
            {
                if (isPixelInside(pixelPositions[i] + offsetPixel)) checkedPositions.Add(pixelPositions[i]);
            }

            return checkedPositions;
        }

        #endregion


    }
}
